import typing
from collections import deque

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.dispatch.handler.func import FuncHandler, Function
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.process import process_inner
from telegrinder.bot.dispatch.return_manager.abc import ABCReturnManager
from telegrinder.bot.dispatch.view.abc import ABCView
from telegrinder.bot.rules.abc import ABCRule, Always
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

type Func[**P, T] = typing.Callable[P, typing.Coroutine[typing.Any, typing.Any, T]]


class BaseView(ABCView):
    def __init__(self, update_type: UpdateType | None = None) -> None:
        self.handlers: deque[ABCHandler] = deque()
        self.middlewares: deque[ABCMiddleware] = deque()
        self.return_manager: ABCReturnManager | None = None
        self.update_type = update_type
        self._auto_rules: ABCRule = Always()

    def __repr__(self) -> str:
        return "<{}>".format(type(self).__name__)

    @property
    def auto_rules(self) -> ABCRule:
        return self._auto_rules

    @auto_rules.setter
    def auto_rules(self, value: ABCRule | list[ABCRule], /) -> None:
        """Example usage:

        ```python
        view.auto_rules = Rule1() & Rule2() | Rule3() & Rule4()
        view.auto_rules # <OrRule>

        view.auto_rules = [Rule1(), Rule2()]
        view.auto_rules # <AndRule>
        ```
        """
        if isinstance(value, list):
            for rule in value:
                self._auto_rules = self._auto_rules & rule
        else:
            self._auto_rules = value

    @classmethod
    def to_handler[T: Function](cls, *rules: ABCRule, final: bool = True) -> typing.Callable[[T], FuncHandler[T]]:
        def wrapper(func: T, /) -> FuncHandler[T]:
            return FuncHandler(
                function=func,
                rules=list(rules),
                final=final,
            )

        return wrapper

    def __call__[T: Function](self, *rules: ABCRule, final: bool = True) -> typing.Callable[[T], T]:
        def wrapper(func: T, /) -> T:
            self.handlers.append(
                FuncHandler(
                    function=func,
                    rules=[self.auto_rules, *rules],
                    final=final,
                ),
            )
            return func

        return wrapper

    @typing.overload
    def register_middleware[T: ABCMiddleware](self, middleware_cls: type[T], /) -> type[T]: ...

    @typing.overload
    def register_middleware(self, middleware: ABCMiddleware, /) -> None: ...

    def register_middleware(
        self,
        middleware: type[ABCMiddleware] | ABCMiddleware,
    ) -> typing.Callable[..., typing.Any] | None:
        self.middlewares.append(middleware() if isinstance(middleware, type) else middleware)
        return middleware if isinstance(middleware, type) else None

    async def check(self, event: Update) -> bool:
        if self.update_type is not None and event.update_type != self.update_type:
            return False
        return bool(self.handlers or self.middlewares)

    async def process(self, event: Update, api: API, context: Context) -> bool:
        return await process_inner(
            api,
            event,
            context,
            self,
        )

    def load(self, external: typing.Self, /) -> None:
        self.handlers.extend(external.handlers)
        self.middlewares.extend(external.middlewares)


__all__ = ("BaseView",)



import asyncio
import contextlib
import datetime
import enum
import typing

from telegrinder.modules import logger
from telegrinder.tools.aio import cancel_future, run_task
from telegrinder.tools.final import Final
from telegrinder.tools.fullname import fullname
from telegrinder.tools.lifespan import (
    CoroutineFunc,
    CoroutineTask,
    DelayedTask,
    Lifespan,
    Task,
    to_coroutine_task,
)
from telegrinder.tools.singleton.singleton import Singleton

type Tasks = set[asyncio.Task[typing.Any]]
type LoopFactory = typing.Callable[[], asyncio.AbstractEventLoop]
type DelayedFunctionDecorator[**P, R] = typing.Callable[[typing.Callable[P, R]], DelayedFunction[P, R]]


class DelayedFunction[**P, R](typing.Protocol):
    __name__: str
    __delayed_task__: DelayedTask

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> typing.Coroutine[typing.Any, typing.Any, R]: ...

    def cancel(self) -> bool:
        """Cancel delayed task."""
        ...


@enum.unique
class LoopWrapperState(enum.Enum):
    NOT_RUNNING = enum.auto()
    RUNNING = enum.auto()
    RUNNING_MANUALLY = enum.auto()
    SHUTDOWN = enum.auto()


@typing.final
class LoopWrapper(Singleton, Final):
    _loop: asyncio.AbstractEventLoop
    _lifespan: Lifespan
    _tasks: list[CoroutineTask[typing.Any]]
    _state: LoopWrapperState
    _all_tasks: set[asyncio.Task[typing.Any]]
    _limit: int | None
    _semaphore: asyncio.Semaphore | None

    __slots__ = (
        "_loop",
        "_lifespan",
        "_tasks",
        "_state",
        "_all_tasks",
        "_limit",
        "_semaphore",
    )

    def __init__(self) -> None:
        self._loop = asyncio.get_event_loop()
        self._lifespan = Lifespan()
        self._tasks = list()
        self._state = LoopWrapperState.NOT_RUNNING
        self._all_tasks = set()
        self._limit = None
        self._semaphore = None

        self._create_task(self._run_async_event_loop())

    def __call__(self) -> asyncio.AbstractEventLoop:
        """A loop factory."""
        return self._loop

    def __repr__(self) -> str:
        return "<{}: loop={!r} lifespan={!r}>".format(
            fullname(self) + (" (running)" if self.running else ""),
            self._loop,
            self._lifespan,
        )

    async def _run_async_event_loop(self) -> None:
        if not self.running:
            self._state = LoopWrapperState.RUNNING
            async with self._async_wrap_loop():
                await self._run()

    async def _run(self) -> None:
        logger.debug("Running loop wrapper")

        while self._tasks:
            await self.create_task(self._tasks.pop(0))

        while self.running and (tasks := self._get_all_tasks()):
            await self._process_tasks(tasks)

    async def _process_tasks(self, tasks: Tasks, /) -> None:
        """Processes the given tasks, checks for exceptions, and raises them if any."""

        tasks_results, _ = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
        for task_result in tasks_results:
            try:
                if not task_result.cancelled() and (exception := task_result.exception()) is not None:
                    raise exception from None  # Raise the exception that was set on the task.
            except BaseException:
                logger.exception("Traceback message below:")

    async def _cancel_tasks(self) -> None:
        with contextlib.suppress(asyncio.CancelledError, asyncio.InvalidStateError):
            await cancel_future(asyncio.gather(*self._get_all_tasks(), return_exceptions=True))

    @contextlib.asynccontextmanager
    async def _async_wrap_loop(self) -> typing.AsyncGenerator[typing.Any, None]:
        try:
            await self._lifespan._start()
            yield
        except asyncio.CancelledError:
            logger.info("LoopWrapper task was cancelled, cancelling tasks...")
            await self._cancel_tasks()
        finally:
            await self._shutdown()

    async def _shutdown(self) -> None:
        await self.lifespan._shutdown()
        logger.debug("Shutting down loop wrapper")
        self._state = LoopWrapperState.SHUTDOWN

    async def _run_coro_with_semaphore(self, coro: CoroutineTask[typing.Any], /) -> None:
        assert self._semaphore is not None
        try:
            await coro
        finally:
            self._semaphore.release()

    async def _create_task_with_semaphore(self, coro: CoroutineTask[typing.Any], /) -> None:
        assert self._semaphore is not None
        await self._semaphore.acquire()
        self._create_task(self._run_coro_with_semaphore(coro))

    def _create_task(self, coro: CoroutineTask[typing.Any], /) -> None:
        task = self._loop.create_task(coro)
        self._all_tasks.add(task)
        task.add_done_callback(self._all_tasks.discard)

    def _get_all_tasks(self) -> Tasks:
        """Get a set of all tasks from the loop wrapper and event loop (`exclude the current task if any`)."""

        return (self._all_tasks | asyncio.all_tasks(loop=self._loop)).symmetric_difference(
            set() if (task := asyncio.current_task(self.loop)) is None else {task},
        )

    def _close_loop(self) -> None:
        if not self._loop.is_closed():
            logger.debug("Closing event loop {!r}", self._loop)
            self._loop.close()

    @contextlib.contextmanager
    def _wrap_loop(self, *, close_loop: bool = True) -> typing.Generator[typing.Any, None, None]:
        try:
            self.lifespan.start()
            yield
        except KeyboardInterrupt:
            logger.info("Caught KeyboardInterrupt, cancelling tasks...")
            run_task(self._cancel_tasks())
        finally:
            run_task(self._shutdown())
            if close_loop:
                self._close_loop()

    def _get_delayed_task_decorator(
        self,
        repeat: bool,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float | datetime.timedelta = 0.0,
    ) -> typing.Callable[..., typing.Any]:
        if isinstance(seconds, int | float):
            seconds += minutes * 60
            seconds += hours * 60 * 60
            seconds += days * 24 * 60 * 60

        def decorator[Func: CoroutineFunc[..., typing.Any]](function: Func) -> Func:
            self.add_task(DelayedTask(function, seconds, repeat=repeat))
            return function

        return decorator

    @property
    def lifespan(self) -> Lifespan:
        return self._lifespan

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def running(self) -> bool:
        return self._state in {LoopWrapperState.RUNNING, LoopWrapperState.RUNNING_MANUALLY}

    @property
    def shutdown(self) -> bool:
        return self._state is LoopWrapperState.SHUTDOWN

    @property
    def tasks_limit(self) -> int | None:
        return self._limit

    def run(self, *, close_loop: bool = True) -> typing.NoReturn:  # type: ignore
        if self.running:
            raise RuntimeError("Loop wrapper already running.")

        self._state = LoopWrapperState.RUNNING_MANUALLY
        with self._wrap_loop(close_loop=close_loop):
            run_task(self._run())

    @typing.overload
    def bind_loop(self, *, loop_factory: LoopFactory) -> typing.Self: ...

    @typing.overload
    def bind_loop(self, *, loop: asyncio.AbstractEventLoop) -> typing.Self: ...

    def bind_loop(
        self,
        *,
        loop_factory: LoopFactory | None = None,
        loop: asyncio.AbstractEventLoop | None = None,
    ) -> typing.Self:
        assert loop is not None or loop_factory is not None

        if self.running:
            logger.warning("Cannot bind a new event loop to running loop wrapper.")
            return self

        old_loop = self._loop
        self._loop = loop_factory() if loop_factory else loop or self._loop

        if old_loop is not self._loop:
            self._create_task(self._run_async_event_loop())

        return self

    def limit(self, value: int, /) -> typing.Self:
        if self._limit is not None:
            raise ValueError("Cannot reset limit value.")

        self._limit = value
        self._semaphore = asyncio.Semaphore(value)
        return self

    def add_task(self, task: Task[..., typing.Any], /) -> None:
        coro_task = to_coroutine_task(task)
        return self._create_task(coro_task) if self.running else self._tasks.append(coro_task)

    async def create_task(self, task: Task[..., typing.Any], /) -> None:
        if not self.running:
            self.add_task(task)
            return

        coro_task = to_coroutine_task(task)
        if self._semaphore is not None:
            await self._create_task_with_semaphore(coro_task)
        else:
            self._create_task(coro_task)

    @typing.overload
    def timer[**P, R](self, delta: datetime.timedelta, /) -> DelayedFunctionDecorator[P, R]: ...

    @typing.overload
    def timer[**P, R](
        self,
        *,
        days: int = ...,
        hours: int = ...,
        minutes: int = ...,
        seconds: float = ...,
    ) -> DelayedFunctionDecorator[P, R]: ...

    def timer(
        self,
        delta: datetime.timedelta | None = None,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0.0,
    ) -> DelayedFunctionDecorator[..., typing.Any]:
        return self._get_delayed_task_decorator(
            repeat=False,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=delta or seconds,
        )

    @typing.overload
    def interval[**P, R](self, delta: datetime.timedelta, /) -> DelayedFunctionDecorator[P, R]: ...

    @typing.overload
    def interval[**P, R](
        self,
        *,
        days: int = ...,
        hours: int = ...,
        minutes: int = ...,
        seconds: float = ...,
    ) -> DelayedFunctionDecorator[P, R]: ...

    def interval(
        self,
        delta: datetime.timedelta | None = None,
        *,
        days: int = 0,
        hours: int = 0,
        minutes: int = 0,
        seconds: float = 0.0,
    ) -> DelayedFunctionDecorator[..., typing.Any]:
        return self._get_delayed_task_decorator(
            repeat=True,
            days=days,
            hours=hours,
            minutes=minutes,
            seconds=delta or seconds,
        )


__all__ = ("DelayedTask", "LoopWrapper", "to_coroutine_task")



from telegrinder import API, Telegrinder, Token
from telegrinder.bot.dispatch.handler.message_reply import MessageReplyHandler
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


bot.on.message.handlers.extend(
    [
        MessageReplyHandler("Good morning!", Text("/gm")),
        MessageReplyHandler("What do you say?", as_reply=True),
    ]
)

bot.run_forever()



from __future__ import annotations

import typing

from telegrinder.node.scope import NodeScope
from telegrinder.tools.global_context import GlobalContext, ctx_var
from telegrinder.tools.global_context.builtin_context import TelegrinderContext

if typing.TYPE_CHECKING:
    from telegrinder.node.base import IsNode
    from telegrinder.node.composer import NodeSession

TELEGRINDER_CONTEXT: typing.Final[TelegrinderContext] = TelegrinderContext()


class NodeGlobalContext(GlobalContext):
    __ctx_name__ = "node_context"

    global_sessions: dict[IsNode, NodeSession] = ctx_var(
        const=True,
        init=False,
        default_factory=dict,
    )

    async def close_global_scopes(self) -> None:
        for session in self.global_sessions.values():
            await session.close(scopes=(NodeScope.GLOBAL,))

        self.global_sessions.clear()


def get_global_session(node: IsNode, /) -> NodeSession | None:
    return NODE_CONTEXT.global_sessions.get(node)


def set_global_session(node: IsNode, session: NodeSession, /) -> None:
    NODE_CONTEXT.global_sessions[node] = session


@TELEGRINDER_CONTEXT.loop_wrapper.lifespan.on_shutdown
async def close_nodes_global_scopes() -> None:
    await NODE_CONTEXT.close_global_scopes()


NODE_CONTEXT: typing.Final[NodeGlobalContext] = NodeGlobalContext()


__all__ = ("NODE_CONTEXT", "NodeGlobalContext", "get_global_session", "set_global_session")



import pytest

from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.middleware.abc import ABCMiddleware
from telegrinder.bot.dispatch.return_manager.abc import register_manager
from telegrinder.bot.dispatch.return_manager.message import MessageReturnManager
from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.bot.rules.abc import ABCRule, AndRule
from telegrinder.bot.rules.text import Text


class CustomMessageReturnManager(MessageReturnManager):
    @register_manager(int)
    @staticmethod
    async def int_manager(value: int) -> None:
        assert isinstance(value, int)


class CustomMessageView(BaseView):
    def __init__(self) -> None:
        super().__init__()
        self.return_manager: CustomMessageReturnManager = CustomMessageReturnManager()


class PreMiddleware(ABCMiddleware):
    def pre(self, event: MessageCute, ctx: Context) -> bool:
        return True


class PostMiddleware(ABCMiddleware):
    def post(self, ctx: Context) -> None:
        assert ctx.responses == [b"123data"]


@pytest.mark.asyncio()
async def test_register_middleware():
    view = CustomMessageView()

    @view.register_middleware
    class SomeMiddleware(ABCMiddleware):
        async def pre(self, event: MessageCute) -> None:
            pass

    assert len(view.middlewares) == 1
    assert isinstance(view.middlewares[0], SomeMiddleware)


@pytest.mark.asyncio()
async def test_register_func_handler():
    view = CustomMessageView()

    @view()
    async def func_handler(event: MessageCute) -> None:
        pass

    assert len(view.handlers) == 1
    assert isinstance(view.handlers[0], FuncHandler)
    assert view.handlers[0].function == func_handler


@pytest.mark.asyncio()
async def test_register_auto_rules():
    view = CustomMessageView()

    class Rule(ABCRule):
        async def check(self, event: MessageCute) -> bool: ...

    view.auto_rules = Rule() & Rule()
    assert isinstance(view.auto_rules, AndRule)


@pytest.mark.asyncio()
async def test_register_return_manager():
    view = CustomMessageView()

    @view.return_manager.register_manager(bytes)
    async def manager(value: bytes, event: MessageCute, ctx: Context) -> None: ...

    assert isinstance(view.return_manager, MessageReturnManager)
    assert manager in view.return_manager.managers


@pytest.mark.asyncio()
async def test_view_check_and_process(api_instance, message_update):
    view = CustomMessageView()

    @view()
    async def handler(message: MessageCute):
        assert isinstance(message, MessageCute) is True

    assert await view.check(message_update) is True
    assert await view.process(message_update, api_instance, Context()) is True


@pytest.mark.asyncio()
async def test_process_pre_middleware(api_instance, message_update):
    view = CustomMessageView()
    view.middlewares.append(PreMiddleware())

    @view()
    async def handler(message: MessageCute):
        return None

    assert await view.check(message_update) is True
    assert await view.process(message_update, api_instance, Context()) is True


@pytest.mark.asyncio()
async def test_process_post_middleware(api_instance, message_update):
    view = CustomMessageView()
    view.middlewares.append(PostMiddleware())

    @view()
    async def handler(message: MessageCute):
        return b"123data"

    assert await view.check(message_update) is True
    assert await view.process(message_update, api_instance, Context()) is True


@pytest.mark.asyncio()
async def test_process_with_rule(api_instance, message_update):
    view = CustomMessageView()

    @view(Text("Hello, Telegrinder! btw, laurelang - nice pure logical programming launguage ^_^"))
    async def handler(message: MessageCute):
        return None

    assert await view.check(message_update) is True
    assert await view.process(message_update, api_instance, Context()) is True


@pytest.mark.asyncio()
async def test_process_with_return_manager(api_instance, message_update):
    view = CustomMessageView()

    @view()
    async def handler(message: MessageCute):
        return 123

    assert await view.check(message_update) is True
    assert await view.process(message_update, api_instance, Context()) is True



from __future__ import annotations

import enum
import typing

if typing.TYPE_CHECKING:
    from telegrinder.node.base import Composable

NODE_SCOPE_KEY: typing.Final[str] = "scope"


class NodeScope(enum.Enum):
    GLOBAL = enum.auto()
    PER_EVENT = enum.auto()
    PER_CALL = enum.auto()


PER_EVENT = NodeScope.PER_EVENT
PER_CALL = NodeScope.PER_CALL
GLOBAL = NodeScope.GLOBAL


def per_call[T: Composable[typing.Any]](node: type[T]) -> type[T]:
    setattr(node, "scope", PER_CALL)
    return node


def per_event[T: Composable[typing.Any]](node: type[T]) -> type[T]:
    setattr(node, "scope", PER_EVENT)
    return node


def global_node[T: Composable[typing.Any]](node: type[T]) -> type[T]:
    setattr(node, "scope", GLOBAL)
    return node


def get_scope(node: Composable[typing.Any], /) -> NodeScope:
    return getattr(node, NODE_SCOPE_KEY, PER_EVENT)


__all__ = (
    "GLOBAL",
    "PER_CALL",
    "PER_EVENT",
    "NodeScope",
    "get_scope",
    "global_node",
    "per_call",
    "per_event",
)



import dataclasses
import typing

import pytest

from telegrinder.api.api import API
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules import IsUser, Markup
from telegrinder.node import DataNode, Node, Polymorphic, impl, scalar_node
from telegrinder.node.base import ComposeError, as_node
from telegrinder.node.composer import NodeCollection, compose_nodes
from telegrinder.node.event import EventNode
from telegrinder.node.me import Me
from telegrinder.node.rule import RuleChain
from telegrinder.node.text import Text
from telegrinder.node.tools.generator import generate_node
from telegrinder.types.objects import Message, Update, User
from tests.test_utils import with_mocked_api

GET_ME_RAW_RESPONSE = """
{
  "ok": true,
  "result": {
    "id": 55732494,
    "is_bot": true,
    "first_name": "Cute",
    "username": "cute123_bot",
    "can_join_groups": true,
    "can_read_all_group_messages": false,
    "supports_inline_queries": true
  }
}
"""


class PrettyRuleChain(
    RuleChain[
        Markup("Hello, <framework>! btw, <logical_lang> - nice pure logical programming launguage <smile>"),
        IsUser(),
    ]
):
    framework: str
    logical_lang: str
    smile: str


@scalar_node[int]
class ScalarMorphNode(Polymorphic):
    @impl
    async def impl1(cls, update: UpdateCute) -> int:
        if update.update_id >= 10000:
            return update.update_id
        raise ComposeError("Update id < 10000")


@dataclasses.dataclass(slots=True, frozen=True)
class DataMorphNode(Polymorphic, DataNode):
    update_id: int

    @impl
    async def impl1(cls, update: UpdateCute) -> typing.Self:
        if update.update_id >= 10000:
            return cls(update.update_id)
        raise ComposeError("Update id < 10000")


class StringNode(Node):
    @classmethod
    async def compose(cls, update: UpdateCute) -> str:
        return str(update.update_id)


@pytest.mark.asyncio()
async def test_scalar_morph(api_instance, message_update):
    result = await compose_nodes(
        {"scalar_morph": as_node(ScalarMorphNode)},
        Context(),
        {API: api_instance, Update: message_update},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert result.value.values == {"scalar_morph": 12345}


@pytest.mark.asyncio()
async def test_data_morph(api_instance, message_update):
    result = await compose_nodes(
        {"data_morph": DataMorphNode}, Context(), {API: api_instance, Update: message_update}
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert result.value.values == {"data_morph": DataMorphNode(12345)}


@pytest.mark.asyncio()
async def test_node(api_instance, message_update):
    result = await compose_nodes({"string": StringNode}, Context(), {API: api_instance, Update: message_update})
    assert result
    assert isinstance(result.value, NodeCollection)
    assert result.value.values == {"string": "12345"}


@pytest.mark.asyncio()
async def test_message_cute_as_node(api_instance, message_update):
    result = await compose_nodes(
        {"message": as_node(MessageCute)},
        Context(),
        {API: api_instance, Update: message_update},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert "message" in result.value.values and isinstance(result.value.values["message"], MessageCute)


@pytest.mark.asyncio()
@with_mocked_api(GET_ME_RAW_RESPONSE)
async def test_me_node(api: API):
    result = await compose_nodes({"me": as_node(Me)}, Context(), {API: api})
    assert result
    assert isinstance(result.value, NodeCollection)
    assert "me" in result.value.values and isinstance(result.value.values["me"], User)
    assert result.value.values["me"].id == 55732494


@pytest.mark.asyncio()
async def test_event_node_with_dict(api_instance, message_update):
    result = await compose_nodes(
        {"event": as_node(EventNode[dict[str, typing.Any]])},
        Context(),
        {API: api_instance, Update: message_update, Context: Context()},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert "event" in result.value.values and isinstance(result.value.values["event"], dict)
    assert result.value.values["event"] == message_update.message.unwrap().to_full_dict()


@pytest.mark.asyncio()
async def test_event_node_with_message_object(api_instance, message_update):
    result = await compose_nodes(
        {"event": as_node(EventNode[Message])},
        Context(),
        {API: api_instance, Update: message_update, Context: Context()},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert "event" in result.value.values and isinstance(result.value.values["event"], Message)
    assert result.value.values["event"] == message_update.message.unwrap()


@pytest.mark.asyncio()
async def test_rule_chain(api_instance, message_update):
    result = await compose_nodes(
        {"rule_chain": PrettyRuleChain},
        Context(),
        {API: api_instance, Update: message_update},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert (
        "rule_chain" in result.value.values
        and result.value.values["rule_chain"].__class__.__name__ == "PrettyRuleChain"
    )
    assert dataclasses.is_dataclass(result.value.values["rule_chain"]) and dataclasses.asdict(
        result.value.values["rule_chain"]  # type: ignore
    ) == {
        "framework": "Telegrinder",
        "logical_lang": "laurelang",
        "smile": "^_^",
    }


@pytest.mark.asyncio()
async def test_node_generator(api_instance, message_update):
    MyNode = generate_node(  # noqa: N806
        (as_node(Text),),
        func=lambda text: text
        if text == "Hello, Telegrinder! btw, laurelang - nice pure logical programming launguage ^_^"
        else False,
    )
    result = await compose_nodes(
        {"my_node": MyNode},
        Context(),
        {API: api_instance, Update: message_update},
    )
    assert result
    assert isinstance(result.value, NodeCollection)
    assert "my_node" in result.value.values and isinstance(result.value.values["my_node"], str)



from telegrinder.tools.callback_data_serialization.abc import ABCDataSerializer
from telegrinder.tools.callback_data_serialization.json_ser import JSONSerializer
from telegrinder.tools.callback_data_serialization.msgpack_ser import MsgPackSerializer

__all__ = ("ABCDataSerializer", "JSONSerializer", "MsgPackSerializer")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import HasEntities, IsChat, IsPrivate, MessageEntities
from telegrinder.tools.formatting import HTMLFormatter, mention
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import MessageEntity

api = API(Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


@bot.on.message(IsChat(), MessageEntities(MessageEntityType.MENTION))
async def handler_mention_me(message: Message, message_entities: list[MessageEntity]):
    my_username = (await api.get_me()).map(lambda me: me.username).unwrap()
    if (
        not my_username
        or my_username != message.text.unwrap()[message_entities[0].offset + 1 : message_entities[0].length]
    ):
        return
    await message.delete()
    await message.answer(
        HTMLFormatter("{:bold} don't mention me please!").format(
            mention(
                message.from_user.first_name,
                user_id=message.from_user.id,
            )
        ),
        parse_mode=HTMLFormatter.PARSE_MODE,
    )


@bot.on.message(IsPrivate(), HasEntities())
async def handler_entities(message: Message):
    await message.answer(
        "Nice {}: {}!".format(
            "entities" if len(message.entities.unwrap()) > 1 else "entity",
            ", ".join(map(lambda e: e.type, message.entities.unwrap())),
        )
    )


bot.run_forever()



import typing

import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.api.error import APIError
from telegrinder.model import Model


class APIResponse(Model):
    ok: bool = False
    result: msgspec.Raw = msgspec.Raw(b"")
    error_code: int = 400
    description: str = "Something went wrong"
    parameters: dict[str, typing.Any] = msgspec.field(default_factory=dict[str, typing.Any])

    def to_result(self) -> Result[msgspec.Raw, APIError]:
        if self.ok:
            return Ok(self.result)
        return Error(APIError(code=self.error_code, error=self.description, data=self.parameters))


__all__ = ("APIResponse",)



class ComposeError(BaseException):
    def __init__(self, message: str = "<no error description>", /) -> None:
        self.message = message
        super().__init__(message)


__all__ = ("ComposeError",)



import pytest

from telegrinder.api.api import API, Token

from ..test_utils import MockedHttpClient


@pytest.fixture()
def api_instance():
    return API(Token("123:ABCdef"), http=MockedHttpClient())



import dataclasses
import sys
import types
import typing as _typing
from functools import cached_property

import typing_extensions as typing
from fntypes.option import Nothing, Option, Some

from telegrinder.tools.global_context.global_context import GlobalContext, ctx_var

type TypeParameter = typing.Union[
    typing.TypeVar,
    typing.TypeVarTuple,
    typing.ParamSpec,
    _typing.TypeVar,
    _typing.TypeVarTuple,
    _typing.ParamSpec,
]
type TypeParameters = tuple[TypeParameter, ...]
type SupportsAnnotations = type[typing.Any] | types.ModuleType | typing.Callable[..., typing.Any]

_CACHED_ANNOTATIONS: typing.Final[GlobalContext] = GlobalContext(
    "cached_annotations",
    annotations=ctx_var(default_factory=dict, const=True),
)


def _cache_annotations(obj: SupportsAnnotations, annotations: dict[str, typing.Any], /) -> None:
    _CACHED_ANNOTATIONS.annotations[obj] = annotations


def _get_cached_annotations(obj: SupportsAnnotations, /) -> dict[str, typing.Any] | None:
    return _CACHED_ANNOTATIONS.annotations.get(obj)


@dataclasses.dataclass
class Annotations:
    obj: SupportsAnnotations

    @cached_property
    def forward_ref_parameters(self) -> dict[str, typing.Any]:
        parameters = dict[str, typing.Any](
            is_argument=False,
            is_class=False,
            module=None,
        )

        if isinstance(self.obj, type):
            parameters["is_class"] = True
            parameters["module"] = (
                sys.modules[module] if (module := getattr(self.obj, "__module__", None)) is not None else None
            )
        elif isinstance(self.obj, types.ModuleType):
            parameters["module"] = self.obj
        elif callable(self.obj):
            parameters["is_argument"] = True

        return parameters

    @cached_property
    def generic_parameters(self) -> Option[dict[TypeParameter, typing.Any]]:
        return get_generic_parameters(self.obj)

    @classmethod
    def from_obj(cls, obj: typing.Any, /) -> typing.Self:
        if not isinstance(obj, type | types.ModuleType | typing.Callable):
            obj = type(obj)

        return cls(obj)

    @typing.overload
    def get(
        self,
        *,
        ignore_failed_evals: bool = True,
        allow_return_type: bool = False,
        cache: bool = False,
    ) -> dict[str, typing.Any | typing.ForwardRef]: ...

    @typing.overload
    def get(
        self,
        *,
        exclude_forward_refs: typing.Literal[True],
        ignore_failed_evals: bool = True,
        allow_return_type: bool = False,
        cache: bool = False,
    ) -> dict[str, typing.Any]: ...

    def get(
        self,
        *,
        exclude_forward_refs: bool = False,
        ignore_failed_evals: bool = True,
        allow_return_type: bool = True,
        cache: bool = False,
    ) -> dict[str, typing.Any]:
        if (cached_annotations := _get_cached_annotations(self.obj)) is not None:
            return cached_annotations

        annotations = dict[str, typing.Any]()
        for name, annotation in typing.get_annotations(
            obj=self.obj,
            format=typing.Format.FORWARDREF,
        ).items():
            if isinstance(annotation, str):
                annotation = typing.ForwardRef(annotation, **self.forward_ref_parameters)

            if not isinstance(annotation, typing.ForwardRef):
                annotations[name] = annotation
                continue

            try:
                value = typing.evaluate_forward_ref(
                    forward_ref=annotation,
                    owner=self.obj,
                    format=typing.Format.VALUE,
                )
            except NameError:
                if not ignore_failed_evals:
                    raise

                value = annotation

            if isinstance(value, typing.ForwardRef) and exclude_forward_refs:
                continue

            annotations[name] = value

        if not allow_return_type:
            annotations.pop("return", None)

        if cache:
            _cache_annotations(self.obj, annotations)

        return annotations


def get_generic_parameters(obj: typing.Any, /) -> Option[dict[TypeParameter, typing.Any]]:
    origin_obj = _typing.get_origin(obj)
    args = _typing.get_args(obj)
    parameters: TypeParameters = getattr(origin_obj or obj, "__parameters__")

    if not parameters:
        return Nothing()

    index = 0
    generic_alias_args = dict[TypeParameter, _typing.Any]()

    for parameter in parameters:
        if isinstance(parameter, _typing.TypeVarTuple):
            stop_index = len(args) - index
            generic_alias_args[parameter] = args[index:stop_index]
            index = stop_index
            continue

        arg = args[index] if index < len(args) else None
        generic_alias_args[parameter] = arg
        index += 1

    return Some(generic_alias_args)


__all__ = ("Annotations", "get_generic_parameters")



from ..models.config import (
    Config,
    MethodsParamsLiteralTypesParam,
    ObjectsFieldsLiteralTypesField,
    ObjectsIdByDefaultField,
    TypedDefaultParameter,
    dec_hook,
)
from ..models.schema import (
    MethodParameter,
    MethodSchema,
    Model,
    ObjectField,
    ObjectSchema,
    TelegramBotAPISchema,
)

__all__ = (
    "Config",
    "MethodParameter",
    "MethodSchema",
    "MethodsParamsLiteralTypesParam",
    "Model",
    "ObjectField",
    "ObjectSchema",
    "ObjectsFieldsLiteralTypesField",
    "ObjectsIdByDefaultField",
    "TelegramBotAPISchema",
    "TypedDefaultParameter",
    "dec_hook",
)



from __future__ import annotations

import dataclasses

import typing_extensions as typing
from fntypes.option import Nothing, Option, Some
from vbml.patcher.abc import ABCPatcher

from telegrinder.api.api import API
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler, Function
from telegrinder.bot.dispatch.middleware.abc import run_middleware
from telegrinder.bot.dispatch.middleware.global_middleware import GlobalMiddleware
from telegrinder.bot.dispatch.view.abc import ABCView
from telegrinder.bot.dispatch.view.box import (
    CallbackQueryView,
    ChatJoinRequestView,
    ChatMemberView,
    ErrorView,
    InlineQueryView,
    MessageView,
    PreCheckoutQueryView,
    RawEventView,
    ViewBox,
)
from telegrinder.modules import logger
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, NodeScope
from telegrinder.tools.global_context import TelegrinderContext
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute
    from telegrinder.bot.rules.abc import ABCRule
    from telegrinder.node.composer import Composer

T = typing.TypeVar("T", default=typing.Any)
R = typing.TypeVar("R", covariant=True, default=typing.Any)
Event = typing.TypeVar("Event", bound="BaseCute")
P = typing.ParamSpec("P", default=...)


@dataclasses.dataclass(repr=False, kw_only=True)
class Dispatch(
    ABCDispatch,
    ViewBox[
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
        ErrorView,
    ],
    typing.Generic[
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        PreCheckoutQueryView,
        RawEventView,
        ErrorView,
    ],
):
    global_context: TelegrinderContext = dataclasses.field(
        init=False,
        default_factory=TelegrinderContext,
    )
    global_middleware: "GlobalMiddleware" = dataclasses.field(
        default_factory=lambda: GlobalMiddleware(),
    )

    def __repr__(self) -> str:
        return "Dispatch(%s)" % ", ".join(f"{k}={v!r}" for k, v in self.get_views().items())

    @property
    def patcher(self) -> ABCPatcher:
        """Alias `patcher` to get `vbml.Patcher` from the global context."""
        return self.global_context.vbml_patcher

    @property
    def composer(self) -> Composer:
        """Alias `composer` to get `telegrinder.node.composer.Composer` from the global context."""
        return self.global_context.composer.unwrap()

    def handle[T: Function](self, *rules: ABCRule, final: bool = True) -> typing.Callable[[T], T]:
        def wrapper(func: T, /) -> T:
            self.raw_event.handlers.append(
                FuncHandler(
                    function=func,
                    rules=list(rules),
                    final=final,
                ),
            )
            return func

        return wrapper

    async def feed(self, event: Update, api: API) -> bool:
        logger.info("New Update(id={}, type={!r})", event.update_id, event.update_type)
        processed = False
        context = Context().add_update_cute(event, api)
        start_time = self.global_context.loop_wrapper.loop.time()

        if (
            await run_middleware(
                self.global_middleware.pre,
                api,
                event,
                context,
                required_nodes=self.global_middleware.pre_required_nodes,
            )
            is False
        ):
            return processed

        for view in self.get_views().values():
            if await view.check(event):
                logger.debug(
                    "Processing update (id={}, type={!r}) with view {!r} by bot (id={})",
                    event.update_id,
                    event.update_type,
                    view,
                    api.id,
                )

                try:
                    if await view.process(event, api, context):
                        processed = True
                        break
                except BaseException as exception:
                    if not await self.error.process(event, api, context.add_exception_update(exception)):
                        raise exception
                finally:
                    for session in context.get(CONTEXT_STORE_NODES_KEY, {}).values():
                        await session.close(scopes=(NodeScope.PER_EVENT,))

        await run_middleware(
            self.global_middleware.post,
            api,
            event,
            context,
            required_nodes=self.global_middleware.post_required_nodes,
        )
        logger.debug(
            "Update (id={}, type={!r}) processed in {} ms by bot (id={})",
            event.update_id,
            event.update_type,
            int((self.global_context.loop_wrapper.loop.time() - start_time) * 1000),
            api.id,
        )
        return processed

    def load(self, external: typing.Self) -> None:
        views_external = external.get_views()

        for name, view in self.get_views().items():
            assert name in views_external, f"View {name!r} is undefined in external dispatch."
            view.load(views_external[name])
            setattr(external, name, view)

        self.global_middleware.filters.difference_update(external.global_middleware.filters)

    def get_view(self, of_type: type[T]) -> Option[T]:
        for view in self.get_views().values():
            if isinstance(view, of_type):
                return Some(view)
        return Nothing()

    def get_views(self) -> dict[str, ABCView]:
        """Get all views."""
        return {
            name: view for name, view in self.__dict__.items() if isinstance(view, ABCView) and name != "error"
        }

    __call__ = handle


__all__ = ("Dispatch",)



from telegrinder.tools.global_context.abc import ABCGlobalContext, CtxVar, GlobalCtxVar
from telegrinder.tools.global_context.builtin_context import TelegrinderContext
from telegrinder.tools.global_context.global_context import GlobalContext, ctx_var, runtime_init

__all__ = (
    "ABCGlobalContext",
    "CtxVar",
    "GlobalContext",
    "GlobalCtxVar",
    "TelegrinderContext",
    "ctx_var",
    "runtime_init",
)



import time

from fntypes.result import Error, Ok

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import ABCRule, IsChat, Markup, Text
from telegrinder.types import ChatPermissions

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


class WithReply(ABCRule):
    async def check(self, message: Message) -> bool:
        if not message.reply_to_message:
            await message.reply("You need to reply to someone's message")
            return False
        return True


class IsChatAdmin(ABCRule, requires=[IsChat()]):
    async def check(self, message: Message) -> bool:
        admins = (await bot.api.get_chat_administrators(chat_id=message.chat.id)).unwrap()
        if message.from_user.id not in (admin.v.user.id for admin in admins if admin.v.user):
            await message.reply("You need to be an admin in this chat")
            return False
        return True


@bot.on.message(Text("/ping"))
async def ping(msg: Message):
    await msg.answer("Pong")


@bot.on.message(Markup(["/mute", "/mute <hours:int>"]), IsChatAdmin(), WithReply())
async def ban(msg: Message, hours: int = 1) -> None:
    perms = ChatPermissions()  # no permissions added
    result = await bot.api.restrict_chat_member(
        chat_id=msg.chat.id,
        user_id=msg.reply_to_message.unwrap().from_user.id,
        permissions=perms,
        until_date=int(time.time() + hours * 3600),
    )
    match result:
        case Ok(_):
            await msg.reply("Done")
        case Error(e):
            await msg.reply("Something went wrong (code {})".format(e.code))


bot.run_forever(skip_updates=True)



import typing

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Argument, Command

T = typing.TypeVar("T")

api = API(Token.from_env())
bot = Telegrinder(api)

logger.set_level("INFO")


def int_validator(value: str) -> int | None:
    if value.isdigit():
        return int(value)
    return None


@bot.on.message(Command("get", Argument("count", validators=[int_validator])))
async def command_handler_get(count: int) -> str:
    return f"You got {count} items."


@bot.on.message(Command("me"))
async def command_handler_me(message: Message) -> list[str]:
    return [
        "First name:",
        message.from_user.first_name,
        "Last name:",
        message.from_user.last_name.unwrap_or("Unknown"),
        "ID:",
        str(message.from_user.id),
    ]


bot.run_forever()



import dataclasses
import typing

from fntypes.result import Error, Ok

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.pre_checkout_query import PreCheckoutQueryCute
from telegrinder.node.base import ComposeError, DataNode, FactoryNode, GlobalNode, scalar_node
from telegrinder.node.polymorphic import Polymorphic, impl
from telegrinder.tools.callback_data_serialization import ABCDataSerializer, JSONSerializer


@scalar_node[str]
class Payload(Polymorphic):
    @impl
    def compose_pre_checkout_query(cls, event: PreCheckoutQueryCute) -> str:
        return event.invoice_payload

    @impl
    def compose_callback_query(cls, event: CallbackQueryCute) -> str:
        return event.data.expect("CallbackQuery has no data.")

    @impl
    def compose_message(cls, event: MessageCute) -> str:
        return event.successful_payment.map(
            lambda payment: payment.invoice_payload,
        ).expect("Message has no successful payment.")


@dataclasses.dataclass(frozen=True, slots=True)
class PayloadSerializer[T: type[ABCDataSerializer[typing.Any]]](DataNode, GlobalNode[T]):
    serializer: type[ABCDataSerializer[typing.Any]]

    @classmethod
    def compose(cls) -> typing.Self:
        return cls(serializer=cls.get(default=JSONSerializer))


class _PayloadData(FactoryNode):
    data_type: type[typing.Any]
    serializer: type[ABCDataSerializer[typing.Any]] | None = None

    def __class_getitem__(
        cls,
        data_type: type[typing.Any] | tuple[type[typing.Any], type[ABCDataSerializer[typing.Any]]],
        /,
    ):
        data_type, serializer = (data_type, None) if not isinstance(data_type, tuple) else data_type
        return cls(data_type=data_type, serializer=serializer)

    @classmethod
    def compose(cls, payload: Payload, payload_serializer: PayloadSerializer) -> typing.Any:
        serializer = cls.serializer or payload_serializer.serializer
        match serializer(cls.data_type).deserialize(payload):
            case Ok(value):
                return value
            case Error(err):
                raise ComposeError(err)


if typing.TYPE_CHECKING:
    import typing_extensions

    DataType = typing.TypeVar("DataType")
    Serializer = typing_extensions.TypeVar(
        "Serializer",
        bound=ABCDataSerializer,
        default=JSONSerializer[typing.Any],
    )

    type PayloadDataType[DataType, Serializer] = typing.Annotated[DataType, Serializer]
    PayloadData: typing.TypeAlias = PayloadDataType[DataType, Serializer]
else:
    PayloadData = _PayloadData


__all__ = ("Payload", "PayloadData", "PayloadSerializer")



import pathlib

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import IsInteger, IsPrivate
from telegrinder.tools.formatting import HTMLFormatter, link

api = API(Token.from_env())
bot = Telegrinder(api)
glossary = pathlib.Path("examples/assets/kaktovik.txt").read_text(encoding="UTF-8")

logger.set_level("INFO")


@bot.on.message(IsInteger())
async def integer_handler(message: Message) -> None:
    integer = int(message.text.unwrap())
    lst = []

    while integer >= 20:
        lst.append(glossary[integer % 20])
        integer = integer // 20

    lst.append(glossary[integer])

    await message.answer("".join(lst[::-1]))


@bot.on.message(IsPrivate())
async def hello_handler() -> HTMLFormatter:
    return HTMLFormatter(
        "Write a positive number and I'll translate it to {} numeral",
    ).format(link("https://en.wikipedia.org/wiki/Kaktovik_numerals", text="kaktovik"))


bot.run_forever()



import typing

from telegrinder.bot.dispatch.handler.abc import ABCHandler

from .short_state import ShortState

if typing.TYPE_CHECKING:
    from telegrinder.bot.cute_types.base import BaseCute


class WaiterActions[Event: BaseCute[typing.Any]](typing.TypedDict):
    on_miss: typing.NotRequired[ABCHandler]
    on_drop: typing.NotRequired[typing.Callable[[ShortState], None]]


__all__ = ("WaiterActions",)



from telegrinder.tools.magic.annotations import Annotations, get_generic_parameters
from telegrinder.tools.magic.dictionary import extract, join_dicts
from telegrinder.tools.magic.function import (
    Bundle,
    bundle,
    function_context,
    get_default_args,
    get_func_annotations,
    get_func_parameters,
    resolve_arg_names,
    resolve_kwonly_arg_names,
    resolve_posonly_arg_names,
)
from telegrinder.tools.magic.shortcut import Shortcut, shortcut

__all__ = (
    "Annotations",
    "Bundle",
    "Shortcut",
    "bundle",
    "extract",
    "function_context",
    "get_default_args",
    "get_func_annotations",
    "get_func_parameters",
    "get_generic_parameters",
    "join_dicts",
    "resolve_arg_names",
    "resolve_kwonly_arg_names",
    "resolve_posonly_arg_names",
    "shortcut",
)



from telegrinder import API, Telegrinder, Token
from telegrinder.rules import PaymentInvoiceCurrency
from telegrinder.types.enums import Currency

bot = Telegrinder(API(Token.from_env()))


@bot.on.pre_checkout_query(PaymentInvoiceCurrency(Currency.XTR))
async def handle_invoice_telegram_stars() -> bool:
    return True  # answer success!


bot.run_forever()



import typing

from fntypes.co import Error, Ok, Result

from telegrinder import Dispatch, LoopWrapper, Message, Polling
from telegrinder.api import API, Token
from telegrinder.api.error import APIError
from telegrinder.rules import Command
from telegrinder.types.objects import User


class DummyAPI(API):
    _cached_me: User | None = None

    async def get_me(self, **other: typing.Any) -> Result[User, APIError]:
        if self._cached_me is None:
            match await super().get_me(**other):
                case Ok(user):
                    self._cached_me = user
                case Error(_) as err:
                    return err
        return Ok(self._cached_me)


api = DummyAPI(token=Token.from_env())
polling = Polling(api=api)
lw = LoopWrapper()
dp = Dispatch()


@dp.raw_event(Command("get_me"))
async def message_handler(message: Message) -> None:
    me = (await api.get_me()).unwrap()
    await message.reply(
        "Hello, my name is {}, my id is {}".format(
            me.full_name,
            me.id,
        )
    )


async def run_polling() -> None:
    async for updates in polling.listen():
        for update in updates:
            await lw.create_task(dp.feed(update, api))


lw.add_task(run_polling())
lw.run()



import dataclasses
import pathlib

from telegrinder.types.objects import InputFile


@dataclasses.dataclass
class InputFileDirectory:
    directory: pathlib.Path
    storage: dict[str, InputFile] = dataclasses.field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.storage = self._load_files()

    def _load_files(self) -> dict[str, InputFile]:
        files = {}

        for path in self.directory.rglob("*"):
            if path.is_file():
                relative_path = path.relative_to(self.directory)
                files[str(relative_path)] = InputFile(path.name, path.read_bytes())

        return files

    def get(self, filename: str, /) -> InputFile:
        assert filename in self.storage, f"File {filename!r} not found."
        return self.storage[filename]


__all__ = ("InputFileDirectory",)



from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.text import HasText
from telegrinder.types.enums import MessageEntityType


class HasMention(ABCRule, requires=[HasText()]):
    def check(self, message: MessageCute) -> bool:
        entities = message.entities.unwrap_or_none()
        if not entities:
            return False
        return any(entity.type == MessageEntityType.MENTION for entity in entities)


__all__ = ("HasMention",)



from .fixtures.api_instance import api_instance
from .fixtures.callback_query_update import callback_query_update
from .fixtures.message_update import message_update

__all__ = ("api_instance", "callback_query_update", "message_update")



import typing
from collections import defaultdict
from typing import Any

from fntypes.option import Some
from fntypes.result import Error, Ok, Result

from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import (
    AnyNode,
    ComposeError,
    IsNode,
    Name,
    NodeClass,
    NodeScope,
    as_node,
    unwrap_node,
)
from telegrinder.node.context import get_global_session, set_global_session
from telegrinder.node.scope import get_scope
from telegrinder.node.session import NodeSession
from telegrinder.tools.aio import Generator, maybe_awaitable, next_generator
from telegrinder.tools.fullname import fullname
from telegrinder.tools.global_context.builtin_context import TelegrinderContext
from telegrinder.tools.magic import bundle, get_func_annotations, join_dicts

type ComposeGenerator = Generator[typing.Any, typing.Any, typing.Any]
type Impls = dict[type[typing.Any], type[typing.Any]]

CONTEXT_STORE_NODES_KEY: typing.Final[str] = "_node_ctx"
GLOBAL_VALUE_KEY: typing.Final[str] = "_value"
TELEGRINDER_CONTEXT: typing.Final[TelegrinderContext] = TelegrinderContext()


def get_impls(
    compose_function: typing.Callable[..., typing.Any],
    impls: Impls,
    /,
) -> dict[str, typing.Any]:
    return {
        key: impls[tp]
        for key, annotation in get_func_annotations(compose_function).items()
        if typing.get_origin(annotation) is type
        and (typing.get_origin(tp := typing.get_args(annotation)[0]) or tp) in impls
    }


async def compose_node(
    node: IsNode,
    linked: dict[type[typing.Any], typing.Any],
    data: dict[type[typing.Any], typing.Any] | None = None,
    impls: Impls | None = None,
) -> "NodeSession":
    subnodes = node.get_subnodes()
    compose = bundle(node.compose, join_dicts(subnodes, linked), bundle_kwargs=True)

    if data:
        compose &= bundle(node.compose, data, typebundle=True)

    if impls:
        compose &= bundle(node.compose, get_impls(node.compose, impls))

    result = compose()
    if node.is_generator():
        generator = typing.cast("ComposeGenerator", result)
        value = await next_generator(generator)
    else:
        generator = None
        value = await maybe_awaitable(result)

    return NodeSession(node, value, generator)


async def compose_nodes(
    nodes: typing.Mapping[str, AnyNode],
    ctx: Context,
    data: dict[type[typing.Any], typing.Any] | None = None,
    impls: Impls | None = None,
) -> Result["NodeCollection", ComposeError]:
    impls = impls or TELEGRINDER_CONTEXT.composer.unwrap().selected_impls
    data = {Context: ctx} | (data or {})
    parent_nodes = dict[IsNode, NodeSession]()
    event_nodes: dict[IsNode, NodeSession] = ctx.get_or_set(CONTEXT_STORE_NODES_KEY, {})
    unwrapped_nodes = {(key, n := (as_node(node)), node): unwrap_node(n) for key, node in nodes.items()}

    for (parent_node_name, parent_node_t, parent_original_type), linked_nodes in unwrapped_nodes.items():
        data.update({Name: parent_node_name, NodeClass: parent_original_type})

        local_nodes = LocalNodeCollection()
        subnodes = dict[Any, Any]()

        for node_t in linked_nodes:
            node_t = as_node(node_t)
            scope = get_scope(node_t)

            if scope is NodeScope.PER_EVENT and node_t in event_nodes:
                local_nodes[node_t] = event_nodes[node_t]
                continue

            if scope is NodeScope.GLOBAL and (global_session := get_global_session(node_t)) is not None:
                local_nodes[node_t] = global_session
                continue

            subnodes |= {
                k: session.value for k, session in (local_nodes | event_nodes).items() if k not in subnodes
            }
            try:
                session = await compose_node(node_t, linked=subnodes, data=data, impls=impls)
            except ComposeError as error:
                await local_nodes.close_local_sessions()
                return Error(ComposeError(f"Cannot compose node `{fullname(node_t)}`, error: {error.message!r}"))

            if scope is NodeScope.PER_CALL:
                await local_nodes.set_local_session(node_t, session)
            else:
                local_nodes[node_t] = session

            if scope is NodeScope.PER_EVENT:
                event_nodes[node_t] = session
            elif scope is NodeScope.GLOBAL:
                set_global_session(node_t, session)

        parent_nodes[parent_node_t] = local_nodes[parent_node_t]

    return Ok(NodeCollection(sessions={k: parent_nodes[t] for k, t, _ in unwrapped_nodes}))


class LocalNodeCollection(dict[IsNode, NodeSession]):
    def __init__(self) -> None:
        super().__init__()

    async def set_local_session(self, node: IsNode, session: NodeSession, /) -> NodeSession:
        if node in self:
            await self[node].close()

        self[node] = session
        return session

    async def close_local_sessions(self) -> None:
        for session in self.values():
            await session.close(scopes=(NodeScope.PER_CALL,))


class NodeCollection:
    __slots__ = ("sessions", "_values")

    def __init__(self, sessions: dict[str, NodeSession]) -> None:
        self.sessions = sessions
        self._values: dict[str, typing.Any] = {}

    def __repr__(self) -> str:
        return "<{}: sessions={!r}>".format(type(self).__name__, self.sessions)

    @property
    def values(self) -> dict[str, typing.Any]:
        if self._values.keys() == self.sessions.keys():
            return self._values

        for name, session in self.sessions.items():
            if name not in self._values:
                self._values[name] = session.value

        return self._values

    async def close_all(
        self,
        with_value: typing.Any | None = None,
        scopes: tuple[NodeScope, ...] = (NodeScope.PER_CALL,),
    ) -> None:
        for session in self.sessions.values():
            await session.close(with_value, scopes=scopes)


class Composer:
    impls: dict[type[typing.Any], set[typing.Any]]
    selected_impls: Impls

    def __init__(self) -> None:
        self.impls = defaultdict(set)
        self.selected_impls = dict()

    def __setitem__(self, for_type: typing.Any, impl: type[typing.Any], /) -> None:
        for_type = typing.get_origin(for_type) or for_type

        if for_type not in self.impls:
            raise LookupError(f"No impls defined for type of `{fullname(for_type)}`.")

        if (typing.get_origin(impl) or impl) not in self.impls[for_type]:
            raise LookupError(f"Impl `{fullname(impl)}` is not defined for type of `{fullname(for_type)}`.")

        self.selected_impls[for_type] = impl

    def impl[T](self, for_type: typing.Any, /) -> typing.Callable[[type[T]], type[T]]:
        def decorator(impl: type[T], /) -> type[T]:
            self.impls[typing.get_origin(for_type) or for_type].add(impl)
            return impl

        return decorator

    async def compose_nodes(
        self,
        nodes: typing.Mapping[str, AnyNode],
        ctx: Context,
        data: dict[type[typing.Any], typing.Any] | None = None,
    ) -> Result[NodeCollection, ComposeError]:
        return await compose_nodes(nodes, ctx, data, self.selected_impls)


TELEGRINDER_CONTEXT.composer = Some(Composer())


__all__ = ("Composer", "NodeCollection", "compose_node", "compose_nodes")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger

logger.set_level("INFO")
bot = Telegrinder(API(Token.from_env()))


@bot.on.message()
async def forward_message(message: Message) -> str:
    result = (await message.forward(message.chat_id)).unwrap()
    forward_origin = result.forward_origin.unwrap().v
    text_response = "Forwarded message that was sent on {} from {}: {}".format(
        forward_origin.date.ctime(), forward_origin.type, "{!r}"
    )

    match forward_origin.type:
        case "chat":
            text_response = text_response.format(
                forward_origin.sender_chat,
            )
        case "channel":
            text_response = text_response.format(
                forward_origin.chat.title.unwrap(),
            )
        case "hidden_user":
            text_response = text_response.format(
                forward_origin.sender_user_name,
            )
        case "user":
            text_response = text_response.format(
                forward_origin.sender_user.full_name,
            )

    return text_response


bot.run_forever()



from telegrinder import API, Checkbox, Message, Telegrinder, Token, WaiterMachine
from telegrinder.bot.dispatch.waiter_machine.hasher.callback import CALLBACK_QUERY_FOR_MESSAGE
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)
logger.set_level("INFO")


@bot.on.message(Text("/checkbox"))
async def action(m: Message):
    picked, m_id = await (
        Checkbox(wm, m.chat.id, "Check your checkbox", cancel_text="Cancel", max_in_row=2)
        .add_option("apple", "Apple", "Apple 🍏")
        .add_option("banana", "Banana", "Banana 🍌", is_picked=True)
        .add_option("pear", "Pear", "Pear 🍐")
        .wait(CALLBACK_QUERY_FOR_MESSAGE, m.ctx_api)
    )
    await m.edit(
        text="You picked: {}".format(", ".join(c for c in picked if picked[c])),
        chat_id=m.chat.id,
        message_id=m_id,
    )


bot.run_forever()



import typing

from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.modules import logger
from telegrinder.node.base import ComposeError, Node, get_nodes
from telegrinder.node.composer import CONTEXT_STORE_NODES_KEY, NodeSession, compose_nodes
from telegrinder.node.scope import NodeScope, get_scope
from telegrinder.tools.aio import maybe_awaitable
from telegrinder.tools.fullname import fullname
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

type Impl = type[classmethod]

MORPH_IMPLEMENTATIONS_KEY: typing.Final[str] = "__morph_implementations__"
IMPL_MARK_KEY: typing.Final[str] = "_is_impl"


@typing.cast("typing.Callable[..., Impl]", lambda f: f)
def impl(method: typing.Callable[..., typing.Any]):
    setattr(method, IMPL_MARK_KEY, True)
    return classmethod(method)


def get_polymorphic_implementations(
    cls: type["Polymorphic"],
    /,
) -> list[typing.Callable[typing.Concatenate[type["Polymorphic"], ...], typing.Any]]:
    moprh_impls = getattr(cls, MORPH_IMPLEMENTATIONS_KEY, None)
    if moprh_impls is not None:
        return moprh_impls

    impls = []
    for cls_ in cls.mro():
        impls += [
            obj.__func__
            for obj in vars(cls_).values()
            if isinstance(obj, classmethod) and getattr(obj.__func__, IMPL_MARK_KEY, False)
        ]

    setattr(cls, MORPH_IMPLEMENTATIONS_KEY, impls)
    return impls


class Polymorphic(Node):
    @classmethod
    async def compose(cls, raw_update: Update, update: UpdateCute, context: Context) -> typing.Any:
        scope = get_scope(cls)
        node_ctx = context.get_or_set(CONTEXT_STORE_NODES_KEY, {})
        data = {
            API: update.ctx_api,
            Context: context,
            Update: raw_update,
        }

        for i, impl_ in enumerate(get_polymorphic_implementations(cls)):
            # To determine whether this is a right morph, all subnodes must be resolved
            if scope is NodeScope.PER_EVENT and (cls, i) in node_ctx:
                return node_ctx[(cls, i)].value

            match await compose_nodes(get_nodes(impl_), context, data=data):
                case Ok(col):
                    node_collection = col
                case Error(err):
                    logger.debug(
                        "Impl `{}` composition failed with error: {!r}",
                        fullname(impl_),
                        err,
                    )
                    continue

            result = None

            try:
                result = await maybe_awaitable(
                    bundle(impl_, data, typebundle=True)(
                        cls,
                        **node_collection.values,
                    ),
                )

                if scope is NodeScope.PER_EVENT:
                    node_ctx[(cls, i)] = NodeSession(cls, result)

                return result
            except ComposeError as compose_error:
                logger.debug(
                    "Failed to compose morph impl `{}`, error: {!r}",
                    fullname(impl_),
                    compose_error.message,
                )
            finally:
                await node_collection.close_all(with_value=result)

        raise ComposeError("No implementation found.")


__all__ = ("Polymorphic", "impl")



import typing
from functools import reduce

from telegrinder import API, ABCMiddleware, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Markup, MessageEntities, Text
from telegrinder.tools.formatting.html_formatter import HTMLFormatter, bold, code_inline
from telegrinder.tools.global_context import GlobalContext, ctx_var
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import MessageEntity, User

logger.set_level("INFO")
api = API(token=Token.from_env())
bot = Telegrinder(api)


class ImportantContext(GlobalContext):
    __ctx_name__ = "important_ctx"

    formatting: bool = False
    users: dict[int, User] = ctx_var(default={}, const=True)


global_ctx = ImportantContext()


def formatting_text(*fmt_texts: str) -> dict[str, typing.Any]:
    params = {"text": "", "parse_mode": None}
    if not global_ctx.formatting:
        params["text"] = "".join(map(str, fmt_texts))
        return params
    params["parse_mode"] = HTMLFormatter.PARSE_MODE
    params["text"] = HTMLFormatter("".join(reduce(lambda x, y: x + y, fmt_texts)))
    return params


@bot.dispatch.message.register_middleware
class UserRegistrarMiddleware(ABCMiddleware):
    def pre(self, event: Message) -> bool:
        if event.from_:
            global_ctx.users[event.from_user.id] = event.from_user
        return True


@bot.on.message(Text("/formatting"))
async def formatting() -> dict[str, typing.Any]:
    global_ctx.formatting = not global_ctx.formatting
    return formatting_text("Formatting ", bold("enabled!" if global_ctx.formatting else "disabled!"))


@bot.on.message(
    MessageEntities([MessageEntityType.TEXT_MENTION, MessageEntityType.MENTION, MessageEntityType.URL])
    & Markup(["/get_user @<username>", "/get_user t.me/<username>", "/get_user <username>"])
)
async def get_user_by_username(
    _: Message, username: str, message_entities: list[MessageEntity]
) -> dict[str, typing.Any]:
    if message_entities[0].type == MessageEntityType.TEXT_MENTION:
        user = message_entities[0].user.unwrap()
        global_ctx.users[user.id] = user
    else:
        user = None
        for u in global_ctx.users.values():
            if u.username.unwrap_or_none() == username:
                user = u

    if user is None:
        return formatting_text("User with username ", bold(username), " not found!")
    return formatting_text("User: ", code_inline(repr(user)))


bot.run_forever()



import abc
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.rules.abc import ABCRule


class BaseReplyHandler(FuncHandler, abc.ABC):
    final: bool

    def __init__(
        self,
        *rules: ABCRule,
        final: bool = True,
        as_reply: bool = False,
        preset_context: Context | None = None,
        **default_params: typing.Any,
    ) -> None:
        self.as_reply = as_reply
        self.default_params = default_params
        super().__init__(
            function=self.handle,
            rules=list(rules),
            final=final,
            preset_context=preset_context or Context(),
        )

    @abc.abstractmethod
    async def handle(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        pass


__all__ = ("BaseReplyHandler",)



import typing

from telegrinder.bot.cute_types import ChatJoinRequestCute
from telegrinder.bot.rules.abc import ABCRule

ChatJoinRequest: typing.TypeAlias = ChatJoinRequestCute


class HasInviteLink(ABCRule):
    def check(self, event: ChatJoinRequest) -> bool:
        return bool(event.invite_link)


class InviteLinkName(ABCRule, requires=[HasInviteLink()]):
    def __init__(self, name: str, /) -> None:
        self.name = name

    def check(self, event: ChatJoinRequest) -> bool:
        return event.invite_link.unwrap().name.unwrap_or_none() == self.name


class InviteLinkByCreator(ABCRule, requires=[HasInviteLink()]):
    def __init__(self, creator_id: int, /) -> None:
        self.creator_id = creator_id

    def check(self, event: ChatJoinRequest) -> bool:
        return event.invite_link.unwrap().creator.id == self.creator_id


__all__ = ("HasInviteLink", "InviteLinkByCreator", "InviteLinkName")



import datetime as dt
import typing
from contextlib import contextmanager

import fntypes.co
import msgspec
from fntypes.result import Error, Ok, Result
from fntypes.variative import Variative

from telegrinder.msgspec_utils.abc import SupportsCast
from telegrinder.msgspec_utils.custom_types.datetime import datetime, timedelta
from telegrinder.msgspec_utils.custom_types.enum_meta import BaseEnumMeta
from telegrinder.msgspec_utils.custom_types.literal import _Literal
from telegrinder.msgspec_utils.custom_types.option import Option
from telegrinder.msgspec_utils.tools import (
    bundle,
    fullname,
    get_origin,
    is_common_type,
    type_check,
)

type Context = dict[str, typing.Any]
type DecHook[T] = typing.Callable[typing.Concatenate[type[T], typing.Any, ...], typing.Any]


def option_dec_hook(
    tp: type[Option[typing.Any]],
    obj: typing.Any,
    /,
) -> fntypes.co.Option[typing.Any] | msgspec.UnsetType:
    if obj is msgspec.UNSET:
        return obj

    if obj is None or isinstance(obj, fntypes.co.Nothing):
        return fntypes.co.Nothing()

    (value_type,) = typing.get_args(tp) or (typing.Any,)
    orig_value_type = typing.get_origin(value_type) or value_type
    orig_obj = obj

    if not isinstance(orig_obj, dict | list) and is_common_type(orig_value_type):
        if orig_value_type is Variative:
            obj = value_type(orig_obj)  # type: ignore
            orig_value_type = typing.get_args(value_type)

        if not type_check(orig_obj, orig_value_type):
            raise msgspec.ValidationError(
                f"Expected `{fullname(orig_value_type)}` or `builtins.None`, got `{fullname(orig_obj)}`.",
            )

        return fntypes.co.Some(obj)

    return fntypes.co.Some(decoder.convert(orig_obj, type=value_type))


def variative_dec_hook(tp: type[Variative], obj: typing.Any, /) -> Variative:
    union_types = typing.get_args(tp)

    if isinstance(obj, dict):
        reverse = False
        models_fields_count: dict[type[msgspec.Struct], int] = {
            m: sum(1 for k in obj if k in m.__struct_fields__)
            for m in union_types
            if issubclass(get_origin(m), msgspec.Struct)
        }
        union_types = tuple(t for t in union_types if t not in models_fields_count)

        if len(set(models_fields_count.values())) != len(models_fields_count.values()):
            models_fields_count = {m: len(m.__struct_fields__) for m in models_fields_count}
            reverse = True

        union_types = (
            *sorted(
                models_fields_count,
                key=lambda k: models_fields_count[k],
                reverse=reverse,
            ),
            *union_types,
        )

    if not isinstance(obj, dict | list) and any(is_common_type(t) and type_check(obj, t) for t in union_types):
        return tp(obj)

    for t in union_types:
        match convert(obj, t):
            case Ok(value):
                return tp(value)
            case Error(_):
                continue
            case _ as arg:
                typing.assert_never(arg)

    raise msgspec.ValidationError(
        "Object of type `{}` doesn't belong to `{}[{}]`.".format(
            fullname(obj),
            fullname(Variative),
            ", ".join(fullname(get_origin(x)) for x in union_types),
        )
    )


def datetime_dec_hook(tp: type[datetime], obj: typing.Any, /) -> datetime:
    if isinstance(obj, datetime):
        return obj

    if isinstance(obj, dt.datetime):
        assert issubclass(tp, SupportsCast)
        return tp.cast(obj)

    if isinstance(obj, int | float):
        return tp.fromtimestamp(timestamp=obj)

    raise TypeError(
        "Cannot validate object of type `{}` into `datetime.datetime`".format(
            fullname(obj),
        ),
    )


def timedelta_dec_hook(tp: type[timedelta], obj: typing.Any, /) -> timedelta:
    if isinstance(obj, timedelta):
        return obj

    if isinstance(obj, dt.timedelta):
        assert issubclass(tp, SupportsCast)
        return tp.cast(obj)

    if isinstance(obj, int | float):
        return tp(seconds=obj)

    raise TypeError(
        "Cannot validate object of type `{}` into `datetime.timedelta`".format(
            fullname(obj),
        ),
    )


def convert[T](
    obj: typing.Any,
    t: type[T],
    /,
    *,
    context: Context | None = None,
) -> Result[T, str]:
    try:
        return Ok(decoder.convert(obj, type=t, strict=True, context=context))
    except msgspec.ValidationError:
        return Error(
            "Expected object of type `{}`, got `{}`.".format(
                fullname(t),
                fullname(obj),
            )
        )


def literal_dec_hook(literal: type[_Literal], obj: typing.Any, /) -> typing.Any:
    if obj in literal.__args__:
        return obj

    raise msgspec.ValidationError(
        "Invalid literal value {!r} of either {}.".format(
            obj,
            literal.__args__[0]
            if len(literal.__args__) == 1
            else (", ".join(f"`{x}`" for x in literal.__args__[:-1])) + f" or `{literal.__args__[-1]}`",
        )
    )


class Decoder:
    """Class `Decoder` for `msgspec` module with decode hook
    for objects with the specified type.

    ```
    import enum

    from datetime import datetime as dt

    class Digit(enum.IntEnum):
        ONE = 1
        TWO = 2
        THREE = 3

    decoder = Decoder()
    decoder.dec_hooks[dt] = lambda t, timestamp: t.fromtimestamp(timestamp)

    decoder.dec_hook(dt, 1713354732)  #> datetime.datetime(2024, 4, 17, 14, 52, 12)

    decoder.convert("123", type=int, strict=False)  #> 123
    decoder.convert(1, type=Digit)  #> <Digit.ONE: 1>

    decoder.decode(b'{"digit":3}', type=dict[str, Digit])  #> {'digit': <Digit.THREE: 3>}
    ```
    """

    dec_hooks: dict[typing.Any, DecHook[typing.Any]]
    abstract_dec_hooks: dict[typing.Any, DecHook[typing.Any]]

    def __init__(self) -> None:
        self.dec_hooks = {
            Option: option_dec_hook,
            Variative: variative_dec_hook,
            datetime: datetime_dec_hook,
            timedelta: timedelta_dec_hook,
            fntypes.option.Some: option_dec_hook,
            fntypes.option.Nothing: option_dec_hook,
        }
        self.abstract_dec_hooks = {
            BaseEnumMeta: lambda enum_type, member: enum_type(member),
            _Literal: literal_dec_hook,
        }

    def __repr__(self) -> str:
        return "<{}: dec_hooks={!r}, abstract_dec_hooks={!r}>".format(
            type(self).__name__,
            self.dec_hooks,
            self.abstract_dec_hooks,
        )

    @typing.overload
    def __call__[T](
        self,
        type: type[T],
        context: Context | None = None,
    ) -> typing.ContextManager[msgspec.json.Decoder[T]]: ...

    @typing.overload
    def __call__(
        self,
        type: typing.Any,
        context: Context | None = None,
    ) -> typing.ContextManager[msgspec.json.Decoder[typing.Any]]: ...

    @typing.overload
    def __call__[T](
        self,
        type: type[T],
        *,
        strict: bool = True,
        context: Context | None = None,
    ) -> typing.ContextManager[msgspec.json.Decoder[T]]: ...

    @typing.overload
    def __call__(
        self,
        type: typing.Any,
        *,
        strict: bool = True,
        context: Context | None = None,
    ) -> typing.ContextManager[msgspec.json.Decoder[typing.Any]]: ...

    @contextmanager
    def __call__(
        self,
        type: typing.Any = object,
        *,
        strict: bool = True,
        context: Context | None = None,
    ) -> typing.Generator[msgspec.json.Decoder, typing.Any, None]:
        """Context manager returns an `msgspec.json.Decoder` object with passed the `dec_hook`."""
        yield msgspec.json.Decoder(
            type=typing.Any if type is object else type,
            strict=strict,
            dec_hook=self.dec_hook(context),
        )

    def add_dec_hook[T](self, t: type[T], /) -> typing.Callable[[DecHook[T]], DecHook[T]]:
        def decorator(func: DecHook[T], /) -> DecHook[T]:
            return self.dec_hooks.setdefault(get_origin(t), func)

        return decorator

    def add_abstract_dec_hook[T](self, abstract_type: type[T], /) -> typing.Callable[[DecHook[T]], DecHook[T]]:
        def decorator(func: DecHook[T], /) -> DecHook[T]:
            return self.abstract_dec_hooks.setdefault(get_origin(abstract_type), func)

        return decorator

    def get_abstract_dec_hook(self, subtype: type[typing.Any], /) -> DecHook[typing.Any] | None:
        for abstract, dec_hook in self.abstract_dec_hooks.items():
            if issubclass(subtype, abstract) or issubclass(type(subtype), abstract):
                return dec_hook

        return None

    def dec_hook(self, context: Context | None = None, /) -> DecHook[typing.Any]:
        def inner(tp: typing.Any, obj: typing.Any, /) -> typing.Any:
            origin_type = get_origin(tp)

            if (dec_hook_func := self.dec_hooks.get(origin_type)) is None and (
                dec_hook_func := self.get_abstract_dec_hook(origin_type)
            ) is None:
                raise NotImplementedError(
                    f"Not implemented decode hook for type `{fullname(origin_type)}`. "
                    "You can implement decode hook for this type.",
                )

            return bundle(dec_hook_func, context or {}, start_idx=2)(tp, obj)

        return inner

    def convert[T](
        self,
        obj: object,
        *,
        type: type[T] = dict,
        strict: bool = True,
        from_attributes: bool = False,
        builtin_types: typing.Iterable[type[typing.Any]] | None = None,
        str_keys: bool = False,
        context: Context | None = None,
    ) -> T:
        return msgspec.convert(
            obj,
            type,
            strict=strict,
            from_attributes=from_attributes,
            dec_hook=self.dec_hook(context),
            builtin_types=builtin_types,
            str_keys=str_keys,
        )

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        context: Context | None = None,
    ) -> typing.Any: ...

    @typing.overload
    def decode[T](
        self,
        buf: str | bytes,
        *,
        type: type[T],
        context: Context | None = None,
    ) -> T: ...

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        type: typing.Any,
        context: Context | None = None,
    ) -> typing.Any: ...

    @typing.overload
    def decode[T](
        self,
        buf: str | bytes,
        *,
        type: type[T],
        strict: bool = True,
        context: Context | None = None,
    ) -> T: ...

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        type: typing.Any,
        strict: bool = True,
        context: Context | None = None,
    ) -> typing.Any: ...

    def decode(
        self,
        buf: str | bytes,
        *,
        type: typing.Any = object,
        strict: bool = True,
        context: Context | None = None,
    ) -> typing.Any:
        return msgspec.json.decode(
            buf,
            type=typing.Any if type is object else type,
            strict=strict,
            dec_hook=self.dec_hook(context),
        )


decoder: typing.Final[Decoder] = Decoder()


__all__ = ("Decoder", "convert", "decoder")



from telegrinder import (
    API,
    CALLBACK_QUERY_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    CallbackQueryCute,
    InlineButton,
    InlineKeyboard,
    MessageCute,
    Telegrinder,
    Token,
    WaiterMachine,
)
from telegrinder.rules import Text

CANCEL_MARKUP = InlineKeyboard().add(InlineButton("Cancel", callback_data="cancel")).get_markup()

api = API(Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)


@bot.on.message(Text("/start"))
async def start_handler(message: MessageCute) -> None:
    input_message = (await message.answer("Input or cancel", reply_markup=CANCEL_MARKUP)).unwrap()
    _, event, _ = await wm.wait_many(
        MESSAGE_FROM_USER(message.from_user.id),
        CALLBACK_QUERY_FOR_MESSAGE(input_message.message_id),
    )

    match event:
        case MessageCute():
            await event.answer(f"Message input: {event.text.unwrap_or('no text')}")
        case CallbackQueryCute():
            await event.answer("Cancel flow")


bot.run_forever()



from fntypes.result import Error, Ok

from telegrinder import API, ChatJoinRequest, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import HasInviteLink, IsUser

bot = Telegrinder(API(Token.from_env()))
logger.set_level("DEBUG")


@bot.on.chat_join_request(HasInviteLink(), IsUser())
async def approve_user(request: ChatJoinRequest) -> None:
    match await request.approve():
        case Ok(ok) if ok:
            await request.ctx_api.send_message(
                chat_id=request.chat.id,
                text=f"Welcome to the chat {request.chat.title.unwrap()!r}, {request.from_user.full_name}!",
            )
        case Error(error):
            logger.error(
                "Something went wrong, error: {!r}, user: {}",
                error.error,
                request.from_user.full_name,
            )


bot.run_forever()



import io
import typing
from abc import ABC, abstractmethod

from telegrinder.client.form_data import MultipartFormProto, encode_form_data

type Data = dict[str, typing.Any] | MultipartFormProto
type Files = dict[str, tuple[str, typing.Any]]


class ABCClient(ABC):
    CONNECTION_TIMEOUT_ERRORS: tuple[type[BaseException], ...] = ()
    CLIENT_CONNECTION_ERRORS: tuple[type[BaseException], ...] = ()

    @abstractmethod
    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        pass

    @property
    @abstractmethod
    def timeout(self) -> float:
        pass

    @abstractmethod
    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        pass

    @abstractmethod
    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    @classmethod
    @abstractmethod
    def multipart_form_factory(cls) -> MultipartFormProto:
        pass

    @classmethod
    def get_form(
        cls,
        *,
        data: dict[str, typing.Any],
        files: Files | None = None,
    ) -> MultipartFormProto:
        multipart_form = cls.multipart_form_factory()
        files = files or {}

        for k, v in encode_form_data(data, files).items():
            multipart_form.add_field(k, v)

        for n, (filename, content) in {
            k: (n, io.BytesIO(c) if isinstance(c, bytes) else c) for k, (n, c) in files.items()
        }.items():
            multipart_form.add_field(n, content, filename=filename)

        return multipart_form

    async def __aenter__(self) -> typing.Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException],
        exc_val: typing.Any,
        exc_tb: typing.Any,
    ) -> None:
        await self.close()


__all__ = ("ABCClient",)



from .attachment import (
    Animation,
    Attachment,
    Audio,
    Document,
    Photo,
    SuccessfulPayment,
    Video,
    VideoNote,
    Voice,
)
from .base import (
    Composable,
    ComposeError,
    DataNode,
    FactoryNode,
    GlobalNode,
    IsNode,
    Name,
    Node,
    NodeComposeFunction,
    NodeConvertable,
    NodeProto,
    NodeType,
    as_node,
    is_node,
    scalar_node,
    unwrap_node,
)
from .callback_query import (
    CallbackQueryData,
    CallbackQueryDataJson,
    Field,
)
from .collection import Collection
from .command import CommandInfo
from .composer import NodeCollection, NodeSession, compose_node, compose_nodes
from .container import ContainerNode
from .context import NODE_CONTEXT, NodeGlobalContext
from .either import Either, Optional
from .error import Error
from .event import EventNode
from .file import File, FileId
from .i18n import ABCTranslator, KeySeparator
from .me import Me
from .payload import Payload, PayloadData, PayloadSerializer
from .polymorphic import Polymorphic, impl
from .reply_message import ReplyMessage
from .rule import RuleChain
from .scope import (
    GLOBAL,
    PER_CALL,
    PER_EVENT,
    NodeScope,
    global_node,
    per_call,
    per_event,
)
from .source import ChatId, ChatSource, Locale, Source, UserId, UserSource
from .text import Caption, Text, TextInteger, TextLiteral
from .tools import generate_node
from .utility import TypeArgs

__all__ = (
    "GLOBAL",
    "NODE_CONTEXT",
    "PER_CALL",
    "PER_EVENT",
    "ABCTranslator",
    "Animation",
    "Attachment",
    "Audio",
    "CallbackQueryData",
    "CallbackQueryDataJson",
    "Caption",
    "ChatId",
    "ChatSource",
    "Collection",
    "CommandInfo",
    "Composable",
    "ComposeError",
    "ContainerNode",
    "DataNode",
    "Document",
    "Either",
    "Error",
    "EventNode",
    "FactoryNode",
    "Field",
    "Field",
    "File",
    "FileId",
    "GlobalNode",
    "IsNode",
    "KeySeparator",
    "Locale",
    "Me",
    "Name",
    "Node",
    "NodeCollection",
    "NodeComposeFunction",
    "NodeConvertable",
    "NodeGlobalContext",
    "NodeProto",
    "NodeScope",
    "NodeSession",
    "NodeType",
    "Optional",
    "Payload",
    "PayloadData",
    "PayloadSerializer",
    "Photo",
    "Polymorphic",
    "ReplyMessage",
    "RuleChain",
    "Source",
    "SuccessfulPayment",
    "Text",
    "TextInteger",
    "TextLiteral",
    "TypeArgs",
    "UserId",
    "UserSource",
    "Video",
    "VideoNote",
    "Voice",
    "as_node",
    "compose_node",
    "compose_nodes",
    "generate_node",
    "global_node",
    "impl",
    "is_node",
    "per_call",
    "per_event",
    "scalar_node",
    "unwrap_node",
)



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text
from telegrinder.tools.formatting import HTMLFormatter, bold, italic, mention

api = API(token=Token.from_env())
api.default_params["parse_mode"] = HTMLFormatter.PARSE_MODE

bot = Telegrinder(api)
logger.set_level("INFO")


@bot.on.message(Text("/formatting"))
async def formatting(m: Message):
    await m.answer(HTMLFormatter(bold(italic("bold italic text!"))))
    await m.answer(
        "python library 'telegrinder' - "
        + HTMLFormatter("{:bold} for effective and reliable telegram {:bold+italic} {:underline}!").format(
            "Framework",
            "bot",
            "building",
        ),
    )
    await m.answer("this is " + mention(m.from_user.first_name, user_id=m.from_user.id))


bot.run_forever()



import inspect
import typing

from fntypes.co import Error, Ok, Result, unwrapping

if typing.TYPE_CHECKING:
    from telegrinder.bot.rules.abc import ABCRule

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.modules import logger
from telegrinder.node.base import get_nodes
from telegrinder.node.composer import compose_nodes
from telegrinder.tools.aio import maybe_awaitable, next_generator, stop_generator
from telegrinder.tools.magic.function import bundle
from telegrinder.types.objects import Update

type When = ABCRule
type Handler = typing.Callable[..., typing.Any]
type ActionFunction = typing.Callable[..., ActionFunctionResult]
type ActionFunctionResult = typing.Union[
    typing.AsyncGenerator[typing.Any, typing.Any],
    typing.Awaitable[typing.Any],
    typing.Generator[typing.Any, typing.Any, typing.Any],
    typing.Any,
]


@unwrapping
async def run_action_function[T: Handler](
    func_handler: FuncHandler[T],
    function: ActionFunction,
    api: API,
    update: Update,
    context: Context,
) -> Result[typing.Any, str]:
    data = {API: api, Update: update}
    node_col = (
        (
            await compose_nodes(
                nodes=get_nodes(function),
                ctx=context,
                data=data,
            )
        )
        .map_err(lambda error: error.message)
        .unwrap()
    )

    temp_ctx = context.copy()
    bundle_function = bundle(function, {Context: temp_ctx, **data}, start_idx=0, typebundle=True)
    bundle_function &= bundle(
        function,
        context | ({} if node_col is None else node_col.values),
        start_idx=0,
    )
    result = bundle_function()

    try:
        if inspect.isasyncgen(result) or inspect.isgenerator(result):
            value = await next_generator(result)
            handler_result = await func_handler.run(api, update, context)
            await stop_generator(result, value)
            return handler_result

        await maybe_awaitable(result)
        return await func_handler.run(api, update, context)
    finally:
        context |= temp_ctx
        await node_col.close_all()


def action[T: Handler](
    function: ActionFunction,
    *,
    when: When | None = None,
) -> typing.Callable[[T], T]:
    def decorator(handler: T, /) -> T:
        func_handler = FuncHandler(function=handler)

        async def action_wrapper(api: API, update: Update, context: Context) -> typing.Any:
            if when and not await check_rule(api, when, update, context):
                logger.debug("When action rule `{!r}` failed.", when)
                result = await func_handler.run(api, update, context)
            else:
                result = await run_action_function(func_handler, function, api, update, context)

            match result:
                case Ok(value):
                    return value
                case Error(error):
                    logger.debug(error)
                    return None

        action_wrapper.__name__ = f"<action for {handler.__name__}>"
        action_wrapper.__qualname__ = f"<action for {handler.__qualname__}>"
        return action_wrapper  # type: ignore

    return decorator


__all__ = ("action",)



import typing

import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.modules import json
from telegrinder.msgspec_utils import decoder
from telegrinder.tools.callback_data_serialization.abc import ABCDataSerializer, ModelType

type Json = dict[str, typing.Any] | ModelType


class JSONSerializer[JsonT: Json](ABCDataSerializer[JsonT]):
    @typing.overload
    def __init__(self, model_t: type[JsonT]) -> None: ...

    @typing.overload
    def __init__(self, model_t: type[JsonT], *, ident_key: str | None = ...) -> None: ...

    def __init__(
        self,
        model_t: type[JsonT] = dict[str, typing.Any],
        *,
        ident_key: str | None = None,
    ) -> None:
        self.model_t = model_t
        self.ident_key: str | None = ident_key or getattr(model_t, "__key__", None)

    @classmethod
    def serialize_from_json(cls, data: JsonT, *, ident_key: str | None = None) -> str:
        return cls(data.__class__, ident_key=ident_key).serialize(data)

    @classmethod
    def deserialize_to_json(cls, serialized_data: str, model_t: type[JsonT]) -> Result[JsonT, str]:
        return cls(model_t).deserialize(serialized_data)

    def serialize(self, data: JsonT) -> str:
        return self.key + json.dumps(data)

    def deserialize(self, serialized_data: str) -> Result[JsonT, str]:
        if not serialized_data.startswith(self.key):
            return Error("Data is not corresponding to key.")

        data = serialized_data.removeprefix(self.key)
        try:
            data_obj = json.loads(data)
        except (msgspec.ValidationError, msgspec.DecodeError):
            return Error("Cannot decode json.")

        if not issubclass(self.model_t, dict):
            try:
                return Ok(decoder.convert(data_obj, type=self.model_t))
            except (msgspec.ValidationError, msgspec.DecodeError):
                return Error("Incorrect data.")

        return Ok(data_obj)


__all__ = ("JSONSerializer",)



import typing

from telegrinder.msgspec_utils.decoder import decoder
from telegrinder.msgspec_utils.encoder import encoder


def dumps(o: typing.Any) -> str:
    return encoder.encode(o)


def loads(s: str | bytes) -> typing.Any:
    return decoder.decode(s)


__all__ = ("dumps", "loads")



import typing

from telegrinder.bot.dispatch.view.base import BaseView
from telegrinder.types.enums import UpdateType

type ChatMemberUpdateType = typing.Literal[
    UpdateType.CHAT_MEMBER,
    UpdateType.MY_CHAT_MEMBER,
]


class ChatMemberView(BaseView):
    def __init__(self, update_type: ChatMemberUpdateType) -> None:
        super().__init__(update_type)


__all__ = ("ChatMemberView",)



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.node import Error
from telegrinder.rules import IsUser, Text

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("DEBUG")


@bot.on.message(Text("oops"))
async def oops_handler(m: Message):
    await m.answer("Oh no")
    raise RuntimeError("Wow")


@bot.on.message(Text("woops"))
async def woops_handler(m: Message):
    await m.answer("Huh it seems like smth oops is gonna happen now...")
    raise ValueError("Wow oopsii!")


@bot.on.error(IsUser())
async def error_handler(err: Error[RuntimeError, ValueError], m: Message):
    await m.answer(f"okay..( Something happened: {err.exception}")


bot.run_forever()



import typing

from telegrinder.tools.fullname import fullname


class Final:
    if not typing.TYPE_CHECKING:

        def __new__(cls, *args, **kwargs):
            if cls is Final:
                raise TypeError("Class Final cannot be instantiate.")
            return super().__new__(cls, *args, **kwargs)

    @typing.final
    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        for base in cls.__bases__:
            if base is not Final and issubclass(base, Final):
                raise TypeError(f"Final class `{fullname(base)}` cannot be subclassed.")


__all__ = ("Final",)



from __future__ import annotations

import asyncio
import ssl
import typing

import aiohttp
import certifi
from aiohttp import BasicAuth, ClientSession, TCPConnector
from aiohttp.hdrs import USER_AGENT
from aiohttp.http import SERVER_SOFTWARE

from telegrinder.__meta__ import __version__
from telegrinder.client.abc import ABCClient
from telegrinder.msgspec_utils import json

if typing.TYPE_CHECKING:
    from aiohttp import ClientResponse

type Data = dict[str, typing.Any] | aiohttp.formdata.FormData
type Response = ClientResponse
type Timeout = int | float | aiohttp.ClientTimeout
type Proxy = str | tuple[str, BasicAuth]
type ProxyChain = typing.Iterable[Proxy]
type ProxyType = Proxy | ProxyChain

DEFAULT_TIMEOUT: typing.Final[float] = 30.0
DEFAULT_LIMIT_SIMULTANEOUS_CONNECTIONS: typing.Final[int] = 100
DEFAULT_TTL_DNS_CACHE: typing.Final[int] = 3600
DEFAULT_HEADERS: typing.Final[dict[str, str]] = {USER_AGENT: f"{SERVER_SOFTWARE} telegrinder/{__version__}"}


def _get_client_timeout(timeout: Timeout, /) -> aiohttp.ClientTimeout:
    return timeout if isinstance(timeout, aiohttp.ClientTimeout) else aiohttp.ClientTimeout(total=float(timeout))


def _prepare_proxy_connector(
    proxy: ProxyType,
    /,
) -> tuple[type[TCPConnector], dict[str, typing.Any] | list[typing.Any]]:
    try:
        from aiohttp_socks import ChainProxyConnector, ProxyConnector, ProxyInfo  # type: ignore
        from aiohttp_socks.utils import parse_proxy_url  # type: ignore
    except ImportError:
        raise ImportError(
            "Module `aiohttp-socks` is not installed. You can install as follows: pip install aiohttp-socks "
            'or pip install "telegrinder[socks]"',
        ) from None

    match proxy:
        case str() | (str(), BasicAuth()):
            proxy_chain = (proxy,)
        case _:
            proxy_chain = proxy

    proxy_infos = list()  # type: ignore

    for _proxy in proxy_chain:
        url, basic = (_proxy, None) if isinstance(_proxy, str) else _proxy
        proxy_type, host, port, username, password = parse_proxy_url(url)  # type: ignore

        if basic is not None:
            username, password = basic.login, basic.password  # type: ignore

        proxy_infos.append(
            ProxyInfo(  # type: ignore
                proxy_type,
                host,
                port,
                username,
                password,
                rdns=True,
            ),
        )

    return (
        (
            ProxyConnector,  # type: ignore
            proxy_infos[0]._asdict(),  # type: ignore
        )
        if len(proxy_infos) == 1
        else (ChainProxyConnector, proxy_infos)
    )  # type: ignore


class AiohttpClient(ABCClient):
    """HTTP client based on `aiohttp` module."""

    CONNECTION_TIMEOUT_ERRORS = (aiohttp.client.ServerConnectionError,)
    CLIENT_CONNECTION_ERRORS = (
        aiohttp.client.ClientConnectionError,
        aiohttp.client.ClientConnectorError,
        aiohttp.ClientOSError,
    )

    __slots__ = (
        "session",
        "limit",
        "ttl_dns_cache",
        "session_params",
        "_proxy",
        "_tcp_connector_kwargs",
        "_tcp_connector_class",
    )

    def __init__(
        self,
        session: ClientSession | None = None,
        proxy: ProxyType | None = None,
        timeout: Timeout = DEFAULT_TIMEOUT,
        limit: int = DEFAULT_LIMIT_SIMULTANEOUS_CONNECTIONS,
        ttl_dns_cache: int = DEFAULT_TTL_DNS_CACHE,
        **session_params: typing.Any,
    ) -> None:
        self.session = session
        self.limit = limit
        self.ttl_dns_cache = ttl_dns_cache
        self.session_params = session_params
        self._timeout = _get_client_timeout(timeout)
        self._proxy = proxy
        self._tcp_connector_kwargs = dict[str, typing.Any](
            ssl=ssl.create_default_context(cafile=certifi.where()),
            limit=limit,
            ttl_dns_cache=ttl_dns_cache,
        )
        self._tcp_connector_class: type[TCPConnector] = TCPConnector

        self.session_params.setdefault("headers", DEFAULT_HEADERS)
        self._setup_proxy()

    def __repr__(self) -> str:
        return "<{}: session={!r}, timeout={}, closed={}>".format(
            type(self).__name__,
            self.session,
            self._timeout,
            True if self.session is None else self.session.closed,
        )

    def __del__(self) -> None:
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector._close()
            self.session._connector = None

    @classmethod
    def multipart_form_factory(cls) -> aiohttp.formdata.FormData:
        return aiohttp.formdata.FormData(quote_fields=False)

    @property
    def timeout(self) -> float:
        return float(self._timeout.total or 0.0)

    async def request_raw(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        *,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> Response:
        if not self.session:
            self.session = ClientSession(
                connector=self._tcp_connector_class(**self._tcp_connector_kwargs),
                json_serialize=json.dumps,
                **self.session_params,
            )

        async with self.session.request(
            method=method,
            url=url,
            data=data,
            timeout=self._timeout if timeout is None else _get_client_timeout(timeout),
            **kwargs,
        ) as response:
            await response.read()
            return response

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        *,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any]:
        response = await self.request_raw(url, method, data, timeout=timeout, **kwargs)
        return await response.json(
            encoding="UTF-8",
            loads=json.loads,
            content_type=None,
        )

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        *,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> str:
        response = await self.request_raw(url, method, data, timeout=timeout, **kwargs)
        return await response.text(encoding="UTF-8")

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        *,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, timeout=timeout, **kwargs)
        if response._body is None:
            await response.read()
        return response._body or bytes()

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: Data | None = None,
        *,
        timeout: Timeout | None = None,
        **kwargs: typing.Any,
    ) -> bytes:
        response = await self.request_raw(url, method, data, timeout=timeout, **kwargs)
        return response._body or bytes()

    async def close(self, *, gracefully: bool = True) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

        if gracefully:
            # Wait 250 ms for graceful shutdown SSL connections
            await asyncio.sleep(0.250)

    def _setup_proxy(self) -> None:
        if self._proxy is not None:
            tcp_connector_class, data = _prepare_proxy_connector(self._proxy)
            self._tcp_connector_class = tcp_connector_class
            self._tcp_connector_kwargs.update(data if isinstance(data, dict) else dict(proxy_infos=data))


__all__ = ("AiohttpClient",)



import typing

import pytest
from fntypes.option import Nothing

from telegrinder.tools.global_context import (
    CtxVar,
    GlobalContext,
    GlobalCtxVar,
    ctx_var,
)


def test_init_anonymous_global_ctx():
    ctx = GlobalContext(value=123)
    assert ctx
    assert "value" in ctx


def test_init_with_name_global_ctx():
    ctx = GlobalContext("context", number=1)

    assert ctx
    assert "number" in ctx
    assert ctx.ctx_name != GlobalContext(number=1).ctx_name


def test_get_value_from_global_ctx():
    ctx = GlobalContext(first_name="Alex")
    ctx.country = "Russia"

    assert ctx.first_name == "Alex"
    assert ctx.country == "Russia"
    assert ctx.get("var") == Nothing()
    assert ctx.get_value("country", str) != Nothing()

    with pytest.raises(
        AssertionError,
        match="Context variable value type of 'str' does not correspond to the expected type 'int'.",
    ):
        ctx.get_value("first_name", int)


def test_set_value_in_global_ctx():
    ctx = GlobalContext()
    ctx.counter = {}
    ctx.start_id = CtxVar(20, const=True)
    ctx.colors = GlobalCtxVar(value=("Red", "Green", "Blue"), name="colors", const=True)

    assert "counter" in ctx
    assert "start_id" in ctx
    assert "colors" in ctx
    assert ctx.counter == {}
    assert ctx.start_id == 20
    assert ctx.colors == ("Red", "Green", "Blue")
    assert ctx.get("counter").unwrap().const is False
    assert ctx.get("start_id").unwrap().const is True
    assert ctx.get("colors").unwrap().const is True

    with pytest.raises(TypeError, match="Unable to set variable 'start_id', because it's a constant."):
        ctx.start_id = 0


def test_del_value_from_global_ctx():
    ctx = GlobalContext(ip="123.123.123.123", port=8888, flag=False)
    del ctx.ip
    del ctx.port

    assert "ip" not in ctx
    assert "port" not in ctx
    assert ctx.pop("flag", bool) != Nothing()
    assert not ctx

    ctx.URL = CtxVar("https://google.com", const=True)
    with pytest.raises(TypeError, match="Unable to delete variable 'URL', because it's a constant."):
        del ctx.URL


def test_global_ctx_with_generic():
    IntGlobalContext = GlobalContext[int]  # noqa: N806
    ctx = IntGlobalContext(one=1, two=2)

    assert isinstance(ctx.one, int)
    assert isinstance(ctx.two, int)
    assert ctx.get_value("one", int) is not Nothing
    assert ctx.get_value("two", int) is not Nothing


def test_rename_var_in_global_ctx():
    ctx = GlobalContext(name="Alex", user_id=CtxVar("123", const=True))

    assert ctx.rename("name", "first_name")
    assert "name" not in ctx
    assert "first_name" in ctx
    assert ctx.first_name == "Alex"

    assert not ctx.rename("user_id", "id")
    assert "user_id" in ctx
    assert "id" not in ctx
    assert ctx.user_id == "123"


def test_clear_global_ctx():
    ctx = GlobalContext(temp=[], state=True)
    ctx.clear()
    assert not ctx

    ctx = GlobalContext(machine=CtxVar("Machine", const=True))
    ctx.clear(include_consts=True)
    assert not ctx


def test_copy_global_ctx():
    ctx = GlobalContext(first=1, second=2, third=3)
    new_ctx = ctx.copy()
    del new_ctx.first

    assert id(ctx) != id(new_ctx)
    assert "first" in ctx
    assert "first" not in new_ctx


def test_delete_ctx_from_global_ctx():
    ctx = GlobalContext("my_context", gun="m4a1", name="John")

    assert ctx.delete_ctx()
    del ctx
    assert not GlobalContext("my_context")


def test_create_global_ctx_with_inheritance():
    class LittleContext(GlobalContext):
        __ctx_name__ = "little_ctx"

        is_little: bool
        little_id: int
        FLAGS: typing.Final[list[int]] = ctx_var(default=[], const=True)

    ctx = LittleContext(is_little=False, little_id=666, FLAGS=[23, 44, 32, 55, 12])

    assert ctx.ctx_name == "little_ctx"
    assert isinstance(ctx, GlobalContext)
    assert issubclass(LittleContext, GlobalContext)

    assert "is_little" in ctx
    assert "little_id" in ctx
    assert "FLAGS" in ctx

    assert ctx.get_value("is_little", bool)
    assert ctx.get_value("little_id", int)
    assert ctx.get_value("FLAGS", list[int])



import pathlib

from telegrinder import API, Message, Telegrinder, Token
from telegrinder.rules import ABCRule

bot = Telegrinder(API(Token.from_env()))
photos_path = pathlib.Path("photos")
photos_path.mkdir(exist_ok=True)


class HasPhoto(ABCRule):
    async def check(self, message: Message) -> bool:
        return bool(message.photo.unwrap_or_none())


@bot.on.message(HasPhoto())
async def downloader(message: Message) -> str:
    photo_file = (await message.ctx_api.get_file(file_id=message.photo.unwrap()[-1].file_id)).unwrap()
    photo_path = photo_file.file_path.unwrap()

    path = photos_path / pathlib.Path(photo_path.split("/")[-1])
    path.write_bytes(
        await message.ctx_api.download_file(photo_path),
    )
    return "Photo downloaded!"


bot.run_forever()



import base64
import binascii
import dataclasses
import typing
from collections import deque
from contextlib import suppress
from functools import cached_property

import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.msgspec_utils import decoder, encoder, get_class_annotations
from telegrinder.tools.callback_data_serialization.abc import ABCDataSerializer, ModelType

DESERIALIZE_EXCEPTIONS: typing.Final[set[type[BaseException]]] = {
    msgspec.DecodeError,
    msgspec.ValidationError,
    binascii.Error,
}

try:
    import brotli  # type: ignore

    DESERIALIZE_EXCEPTIONS.add(brotli.error)  # type: ignore
except ImportError:
    brotli = None


@dataclasses.dataclass(frozen=True, slots=True)
class ModelParser[Model: ModelType]:
    model_type: type[Model]

    def _is_model(self, obj: typing.Any, /) -> typing.TypeGuard[ModelType]:
        return dataclasses.is_dataclass(obj) or isinstance(obj, msgspec.Struct)

    def _model_to_dict(self, model: ModelType) -> dict[str, typing.Any]:
        if dataclasses.is_dataclass(model):
            return dataclasses.asdict(model)
        return msgspec.structs.asdict(model)  # type: ignore

    def _is_union(self, inspected_type: msgspec.inspect.Type, /) -> typing.TypeGuard[msgspec.inspect.UnionType]:
        return isinstance(inspected_type, msgspec.inspect.UnionType)

    def _is_model_type(
        self,
        inspected_type: msgspec.inspect.Type,
        /,
    ) -> typing.TypeGuard[msgspec.inspect.DataclassType | msgspec.inspect.StructType]:
        return isinstance(inspected_type, msgspec.inspect.DataclassType | msgspec.inspect.StructType)

    def _is_iter_of_model(
        self,
        inspected_type: msgspec.inspect.Type,
        /,
    ) -> typing.TypeGuard[msgspec.inspect.ListType]:
        return isinstance(
            inspected_type,
            msgspec.inspect.ListType | msgspec.inspect.SetType | msgspec.inspect.FrozenSetType,
        ) and self._is_model_type(inspected_type.item_type)

    def _inspect_annotation(self, annotation: typing.Any, /) -> tuple[type[ModelType], bool] | None:
        is_iter_of_model = False
        type_args: tuple[msgspec.inspect.Type, ...] = tuple()
        inspected_type = msgspec.inspect.type_info(annotation)

        if self._is_union(inspected_type):
            type_args = inspected_type.types
        elif self._is_iter_of_model(inspected_type):
            type_args = (inspected_type.item_type,)
            is_iter_of_model = True
        elif self._is_model_type(inspected_type):
            type_args = (inspected_type,)

        for arg in type_args:
            if self._is_union(arg):
                type_args += arg.types

            if self._is_model_type(arg):
                return (arg.cls, is_iter_of_model)

            if self._is_iter_of_model(arg):
                return (arg.item_type.cls, True)  # type: ignore

        return None

    def parse(self, model: Model) -> list[typing.Any]:
        """Returns a parsed model as linked list."""
        linked: list[typing.Any] = []
        stack: list[typing.Any] = [(list(self._model_to_dict(model).values()), linked)]

        while stack:
            current_obj, current = stack.pop()

            for item in current_obj:
                if self._is_model(item):
                    item = self._model_to_dict(item)
                if isinstance(item, dict):
                    new_list = []
                    current.append(new_list)
                    stack.append((list(item.values()), new_list))
                elif isinstance(item, list | tuple):
                    new_list = []
                    current.append(new_list)
                    stack.append((item, new_list))
                else:
                    current.append(item)

        return encoder.to_builtins(linked)

    def compose(self, linked: list[typing.Any]) -> dict[str, typing.Any]:
        """Compose linked list to dictionary based on the model class annotations `(without validation)`."""
        root_converted_data: dict[str, typing.Any] = {}
        stack: deque[typing.Any] = deque([(linked, self.model_type, root_converted_data)])

        while stack:
            current_data, current_model, converted_data = stack.pop()

            for index, (field, annotation) in enumerate(get_class_annotations(current_model).items()):
                obj, model_type, is_iter_of_model = current_data[index], None, False

                if isinstance(obj, list) and (inspected := self._inspect_annotation(annotation)):
                    model_type, is_iter_of_model = inspected

                if model_type is not None:
                    if is_iter_of_model:
                        converted_data[field] = []
                        for item in obj:
                            new_converted_data = {}
                            converted_data[field].append(new_converted_data)
                            stack.append((item, model_type, new_converted_data))
                    else:
                        new_converted_data = {}
                        converted_data[field] = new_converted_data
                        stack.append((obj, model_type, new_converted_data))
                else:
                    converted_data[field] = obj

        return root_converted_data


class MsgPackSerializer[Model: ModelType](ABCDataSerializer[Model]):
    @typing.overload
    def __init__(self, model_t: type[Model], /) -> None: ...

    @typing.overload
    def __init__(self, model_t: type[Model], /, *, ident_key: str | None = ...) -> None: ...

    def __init__(self, model_t: type[Model], /, *, ident_key: str | None = None) -> None:
        self.model_t = model_t
        self.ident_key: str | None = ident_key or getattr(model_t, "__key__", None)
        self._model_parser = ModelParser(model_t)

    @classmethod
    def serialize_from_model(cls, model: Model, *, ident_key: str | None = None) -> str:
        return cls(model.__class__, ident_key=ident_key).serialize(model)

    @classmethod
    def deserialize_to_json(cls, serialized_data: str, model_t: type[Model]) -> Result[Model, str]:
        return cls(model_t).deserialize(serialized_data)

    @cached_property
    def key(self) -> bytes:
        return msgspec.msgpack.encode(super().key) if self.ident_key else b""

    def serialize(self, data: Model, /) -> str:
        encoded = self.key + msgspec.msgpack.encode(self._model_parser.parse(data), enc_hook=encoder.enc_hook)
        if brotli is not None:
            return base64.b85encode(brotli.compress(encoded, quality=11)).decode()  # type: ignore
        return base64.urlsafe_b64encode(encoded).decode()

    def deserialize(self, serialized_data: str, /) -> Result[Model, str]:
        with suppress(*DESERIALIZE_EXCEPTIONS):
            if brotli is not None:
                ser_data = typing.cast("bytes", brotli.decompress(base64.b85decode(serialized_data)))
            else:
                ser_data = base64.urlsafe_b64decode(serialized_data)

            if not ser_data.startswith(self.key):
                return Error("Data is not corresponding to key.")

            data: list[typing.Any] = msgspec.msgpack.decode(
                ser_data.removeprefix(self.key),
                dec_hook=decoder.dec_hook(),
            )
            return Ok(decoder.convert(self._model_parser.compose(data), type=self.model_t))

        return Error("Incorrect data.")


__all__ = ("MsgPackSerializer",)



from __future__ import annotations

import dataclasses
import typing

from telegrinder.modules import logger
from telegrinder.node.exceptions import ComposeError
from telegrinder.node.scope import NodeScope, get_scope
from telegrinder.tools.aio import Generator, stop_generator
from telegrinder.tools.fullname import fullname

if typing.TYPE_CHECKING:
    from telegrinder.node.base import IsNode


@dataclasses.dataclass(slots=True, repr=False)
class NodeSession:
    node: IsNode
    value: typing.Any
    generator: Generator[typing.Any, typing.Any, typing.Any] | None = None

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.value!r}" + (" (ACTIVE)>" if self.generator else ">")

    async def close(
        self,
        with_value: typing.Any | None = None,
        scopes: tuple[NodeScope, ...] = (NodeScope.PER_CALL,),
    ) -> typing.Any | None:
        result = None

        if self.generator is not None and get_scope(self.node) in scopes:
            try:
                await stop_generator(self.generator, with_value)
            except ComposeError as compose_error:
                logger.debug(
                    "Caught compose error when closing session for node `{}`: {}",
                    fullname(self.node),
                    compose_error.message,
                )
            except BaseException as exception:
                logger.debug(
                    "Uncaught {!r} was occurred when closing session for node `{}`",
                    exception,
                    fullname(self.node),
                )
            finally:
                self.generator = None

        return result


__all__ = ("NodeSession",)



import typing

from telegrinder.api.api import API, Token
from telegrinder.client.abc import ABCClient


class MockedHttpClient(ABCClient):
    def __init__(
        self,
        return_value: typing.Any | None = None,
        callback: typing.Callable | None = None,
    ) -> None:
        super().__init__()
        self.return_value = return_value
        self.callback = callback or (lambda method, url, data: None)

    @property
    def timeout(self) -> float:
        return 0.0

    async def request_text(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs,
    ) -> str | typing.Any:
        return self.return_value or self.callback(method, url, data)

    async def request_json(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> dict[str, typing.Any] | typing.Any:
        return self.return_value or self.callback(method, url, data)

    async def request_bytes(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> bytes | typing.Any:
        return self.return_value or self.callback(method, url, data)

    async def request_content(
        self,
        url: str,
        method: str = "GET",
        data: dict[str, typing.Any] | None = None,
        **kwargs: typing.Any,
    ) -> bytes | typing.Any:
        return self.return_value or self.callback(method, url, data)

    @classmethod
    def multipart_form_factory(cls) -> typing.Any:
        pass

    @classmethod
    def get_form(
        cls,
        data: dict[str, typing.Any],
        files: dict[str, tuple[str, bytes]] | None = None,
    ) -> typing.Any:
        pass

    async def close(self) -> None:
        pass


def with_mocked_api(return_value: typing.Any):
    def decorator(func: typing.Callable):
        async def wrapper(*args, **kwargs):
            api = API(Token("123:ABCdef"), http=MockedHttpClient(return_value))
            return await func(*args, **kwargs, api=api)

        return wrapper

    return decorator



import re
import typing

import vbml

from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.node.text import Caption, Text
from telegrinder.tools.global_context.builtin_context import TelegrinderContext

type PatternLike = str | vbml.Pattern

TELEGRINDER_CONTEXT: typing.Final[TelegrinderContext] = TelegrinderContext()


def check_string(patterns: list[vbml.Pattern], s: str, ctx: Context) -> bool:
    for pattern in patterns:
        match TELEGRINDER_CONTEXT.vbml_patcher.check(pattern, s):
            case {**response}:
                ctx |= response
            case None | False:
                continue

        return True

    return False


class Markup(ABCRule):
    """Markup Language. See the [vbml documentation](https://github.com/tesseradecade/vbml/blob/master/docs/index.md)."""

    def __init__(
        self,
        patterns: PatternLike | list[PatternLike],
        /,
        *,
        flags: re.RegexFlag | None = None,
    ) -> None:
        self.patterns = [
            vbml.Pattern(text=pattern, flags=flags or TELEGRINDER_CONTEXT.vbml_pattern_flags)
            if isinstance(pattern, str)
            else pattern
            for pattern in ([patterns] if not isinstance(patterns, list) else patterns)
        ]

    def check(self, text: Text | Caption, ctx: Context) -> bool:
        return check_string(self.patterns, text, ctx)


__all__ = ("Markup", "check_string")



import typing

from telegrinder.model import Model
from telegrinder.tools.magic.shortcut import shortcut


def compose_method_params[Cute: BaseCute](
    params: dict[str, typing.Any],
    update: Cute,
    *,
    default_params: set[str | tuple[str, str]] | None = None,
    validators: dict[str, typing.Callable[[Cute], bool]] | None = None,
) -> dict[str, typing.Any]:
    default_params = default_params or set()
    validators = validators or {}

    for param in default_params:
        param_name = param if isinstance(param, str) else param[0]
        if param_name not in params:
            if param_name in validators and not validators[param_name](update):
                continue
            params[param_name] = getattr(update, param if isinstance(param, str) else param[1])

    return params


if typing.TYPE_CHECKING:
    from fntypes.option import Option

    from telegrinder.api.api import API
    from telegrinder.node.base import Node
    from telegrinder.types.objects import Update

    class BaseCute[T: Model](Model):
        api: API

        @classmethod
        def as_node(cls) -> type[Node]: ...

        @classmethod
        def from_update(cls, update: T, bound_api: API) -> typing.Self: ...

        @property
        def ctx_api(self) -> API: ...

        @property
        def raw_update(self) -> Update: ...

        def bind_raw_update(self, raw_update: Update, /) -> typing.Self: ...

        def get_raw_update(self) -> Option[Update]: ...

        def to_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]: ...

        def to_full_dict(
            self,
            *,
            exclude_fields: set[str] | None = None,
        ) -> dict[str, typing.Any]: ...

else:
    from fntypes.co import Some, Variative
    from fntypes.misc import from_optional

    from telegrinder.msgspec_utils import Option, encoder, struct_asdict
    from telegrinder.msgspec_utils import get_class_annotations as _get_class_annotations
    from telegrinder.types.objects import Update

    BOUND_API_KEY = "bound_api"
    RAW_UPDATE_BIND_KEY = "raw_update_bind"

    def _get_cute_from_generic(generic_args, /):
        for arg in generic_args:
            orig_arg = typing.get_origin(arg) or arg

            if not isinstance(orig_arg, type):
                continue
            if orig_arg in (Variative, Some, Option):
                return _get_cute_from_generic(typing.get_args(arg))
            if issubclass(orig_arg, BaseCute):
                return arg

        return None

    def _get_cute_annotations(annotations, /):
        cute_annotations = {}

        for key, hint in annotations.items():
            if not isinstance(hint, type):
                if (cute := _get_cute_from_generic(typing.get_args(hint))) is not None:
                    cute_annotations[key] = cute

            elif issubclass(hint, BaseCute):
                cute_annotations[key] = hint

        return cute_annotations

    def _maybe_wrapped(value, /):
        wrapped_value = None

        while isinstance(value, Variative | Some):
            wrapped_value = value

            if isinstance(value, Variative):
                value = value.v

            if isinstance(value, Some):
                value = value.value

        return wrapped_value if wrapped_value is not None else value

    def _wrap_value(value, type_):
        args = [type_]
        types = []

        while args:
            arg = args.pop(0)
            origin_arg = typing.get_origin(arg) or arg
            if issubclass(origin_arg, Variative | Option):
                args.extend(typing.get_args(arg))
                types.append(Some if issubclass(origin_arg, Option) else arg)

        result = value
        for t in types[::-1]:
            result = t(result)
        return result

    def _to_cute(cls, field, value, bound_api):
        maybe_wrapped_value = _maybe_wrapped(value)
        is_wrapped_value = isinstance(maybe_wrapped_value, Variative | Some)
        cute = cls.__cute_annotations__[field].from_update(
            maybe_wrapped_value._value if is_wrapped_value else maybe_wrapped_value,
            bound_api=bound_api,
        )
        return _wrap_value(cute, cls.__annotations__[field]) if is_wrapped_value else cute

    class BaseCute[T]:
        def __init_subclass__(cls, *args, **kwargs):
            cls.__is_resolved_annotations__ = False
            cls.__cute_annotations__ = None
            cls.__event_node__ = None

        @classmethod
        def as_node(cls):
            if cls.__event_node__ is None:
                from telegrinder.node.event import EventNode

                cls.__event_node__ = EventNode[cls]

            return cls.__event_node__

        @classmethod
        def from_update(cls, update, bound_api):
            if not cls.__is_resolved_annotations__:
                cls.__is_resolved_annotations__ = True
                cls.__annotations__ = _get_class_annotations(cls)

            if cls.__cute_annotations__ is None:
                cls.__cute_annotations__ = _get_cute_annotations(cls.__annotations__)

            cute = cls(
                **{
                    field: _to_cute(cls, field, value, bound_api) if field in cls.__cute_annotations__ else value
                    for field, value in update.to_dict().items()
                },
            )._set_bound_api(api=bound_api)
            return cute.bind_raw_update(update) if isinstance(update, Update) else cute

        @property
        def api(self):
            return self.__dict__[BOUND_API_KEY]

        @property
        def raw_update(self):
            return self.get_raw_update().expect(
                ValueError(f"Cute model `{type(self).__name__}` has no bind `Update` object."),
            )

        @property
        def ctx_api(self):
            return self.api

        def get_raw_update(self):
            return from_optional(self.__dict__.get(RAW_UPDATE_BIND_KEY))

        def bind_raw_update(self, raw_update, /):
            cuties = [self]

            if isinstance(self, Update) and isinstance(cute := self.incoming_update, BaseCute):
                cuties.append(cute)

            for cute in cuties:
                cute.__dict__[RAW_UPDATE_BIND_KEY] = raw_update

            return self

        def _set_bound_api(self, api):
            self.__dict__[BOUND_API_KEY] = api
            return self

        def _to_dict(self, dct_name, exclude_fields, full):
            if dct_name not in self.__dict__:
                dct = struct_asdict(self)
                self.__dict__[dct_name] = dct if not full else encoder.to_builtins(dct, order="deterministic")

            if not exclude_fields:
                return self.__dict__[dct_name]

            return {key: value for key, value in self.__dict__[dct_name].items() if key not in exclude_fields}

        def to_dict(self, *, exclude_fields=None, full=False):
            return self._to_dict(
                "model_as_dict" if not full else "model_as_full_dict",
                exclude_fields=exclude_fields or set(),
                full=full,
            )

        def to_full_dict(self, *, exclude_fields=None):
            return self.to_dict(exclude_fields=exclude_fields, full=True)


__all__ = ("BaseCute", "compose_method_params", "shortcut")



from telegrinder.tools.formatting.deep_links.links import *
from telegrinder.tools.formatting.deep_links.parsing import *
from telegrinder.tools.formatting.deep_links.validators import *

__all__ = (
    "NO_VALUE",
    "NoValue",
    "Parameter",
    "get_parameter_metadata",
    "get_query_params",
    "parse_deep_link",
    "parse_query_params",
    "separate_by_plus_char",
    "tg_bot_attach_open_any_chat",
    "tg_bot_attach_open_current_chat",
    "tg_bot_attach_open_specific_chat",
    "tg_bot_start_link",
    "tg_bot_startchannel_link",
    "tg_bot_startgroup_link",
    "tg_chat_folder_link",
    "tg_chat_invite_link",
    "tg_direct_mini_app_link",
    "tg_emoji_link",
    "tg_emoji_stickerset_link",
    "tg_invoice_link",
    "tg_language_pack_link",
    "tg_main_mini_app_link",
    "tg_mention_link",
    "tg_open_message_link",
    "tg_premium_multigift_link",
    "tg_premium_offer_link",
    "tg_private_channel_boost_link",
    "tg_private_message_link",
    "tg_public_channel_boost_link",
    "tg_public_message_link",
    "tg_public_username_link",
    "tg_share_link",
    "tg_story_link",
)



import typing

from fntypes.result import Ok

from telegrinder.api.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.node.base import ComposeError, FactoryNode, IsNode
from telegrinder.node.composer import compose_nodes
from telegrinder.node.scope import per_call
from telegrinder.tools.fullname import fullname
from telegrinder.types.objects import Update


@per_call
class _Either(FactoryNode):
    """Represents a node that either to compose `left` or `right` nodes.

    For example:
    ```python
    # ScalarNode `Integer` -> int
    # ScalarNode `Float` -> float

    Number = Either[Integer, Float]  # using a type alias just as an example

    def number_to_int(number: Number) -> int:
        return int(number)
    ```
    """

    nodes: tuple[IsNode, IsNode | None]

    def __class_getitem__(cls, node: IsNode | tuple[IsNode, IsNode], /) -> typing.Any:
        nodes = (node, None) if not isinstance(node, tuple) else node
        assert len(nodes) == 2, "Node `Either` must have at least two nodes."
        return cls(nodes=nodes)

    @classmethod
    async def compose(cls, api: API, update: Update, context: Context) -> typing.Any:
        composed = True

        for node in cls.nodes:
            if node is not None:
                match await compose_nodes(dict(node_result=node), context, {API: api, Update: update}):
                    case Ok(col):
                        value = col.values["node_result"]
                        yield value
                        await col.close_all(with_value=value)
                        break
                    case _:
                        composed = False
            else:
                yield None
                break

        if not composed:
            raise ComposeError(
                "Cannot compose either nodes: {}.".format(", ".join(fullname(n) for n in cls.nodes))
            )


if typing.TYPE_CHECKING:
    type Either[Left, Right: typing.Any | None] = Left | Right
    type Optional[Left] = Either[Left, None]
else:
    Either = type("Either", (_Either,), {"__module__": __name__})
    Optional = type("Optional", (_Either,), {"__module__": __name__})


__all__ = ("Either", "Optional")



import abc
import typing
from contextlib import suppress

from telegrinder.bot.cute_types import CallbackQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.payload import (
    PayloadEqRule,
    PayloadJsonEqRule,
    PayloadMarkupRule,
    PayloadModelRule,
)
from telegrinder.tools.aio import maybe_awaitable

type Validator = typing.Callable[[typing.Any], bool | typing.Awaitable[bool]]
type CallbackMap = dict[str, typing.Any | type[typing.Any] | Validator | CallbackMap]
type CallbackMapStrict = dict[str, Validator | CallbackMapStrict]

CallbackQuery: typing.TypeAlias = CallbackQueryCute
CallbackDataEq: typing.TypeAlias = PayloadEqRule
CallbackDataJsonEq: typing.TypeAlias = PayloadJsonEqRule
CallbackDataMarkup: typing.TypeAlias = PayloadMarkupRule
CallbackDataJsonModel: typing.TypeAlias = PayloadModelRule


class HasData(ABCRule):
    def check(self, event: CallbackQuery) -> bool:
        return bool(event.data)


class CallbackQueryDataRule(ABCRule, abc.ABC, requires=[HasData()]):
    pass


class CallbackDataMap(CallbackQueryDataRule):
    def __init__(self, mapping: CallbackMap, /, *, allow_extra_fields: bool = False) -> None:
        """Callback data map validation.
        :param mapping: A callback data mapping with validators.
        :param allow_extra_fields: Allows extra fields in a callback query data.
        """
        self.mapping = self.transform_to_callbacks(mapping)
        self.allow_extra_fields = allow_extra_fields

    @classmethod
    def transform_to_callbacks(cls, callback_map: CallbackMap) -> CallbackMapStrict:
        """Transforms `CallbackMap` to `CallbackMapStrict`."""
        callback_map_result = []

        for key, value in callback_map.items():
            if isinstance(value, type):
                validator = (lambda tp: lambda v: isinstance(v, tp))(value)
            elif isinstance(value, dict):
                validator = cls.transform_to_callbacks(value)
            elif not callable(value):
                validator = (lambda val: lambda v: val == v)(value)
            else:
                validator = value
            callback_map_result.append((key, validator))

        return dict(callback_map_result)

    @staticmethod
    async def run_validator(value: typing.Any, validator: Validator) -> bool:
        """Runs sync/async validator."""
        with suppress(BaseException):
            return await maybe_awaitable(validator(value))
        return False

    @classmethod
    async def match(cls, callback_data: dict[str, typing.Any], callback_map: CallbackMapStrict) -> bool:
        """Matches `callback_data` with `callback_map` recursively."""
        for key, validator in callback_map.items():
            if key not in callback_data:
                return False

            if isinstance(validator, dict):
                if not (isinstance(callback_data[key], dict) and await cls.match(callback_data[key], validator)):
                    return False

            elif not await cls.run_validator(callback_data[key], validator):
                return False

        return True

    async def check(self, event: CallbackQuery, ctx: Context) -> bool:
        callback_data = event.decode_data().unwrap_or_none()
        if callback_data is None:
            return False

        if not self.allow_extra_fields and self.mapping.keys() != callback_data.keys():
            return False

        if await self.match(callback_data, self.mapping):
            ctx.update({k: callback_data[k] for k in self.mapping})
            return True

        return False


__all__ = (
    "CallbackDataEq",
    "CallbackDataJsonEq",
    "CallbackDataJsonModel",
    "CallbackDataMap",
    "CallbackDataMarkup",
    "CallbackQueryDataRule",
    "HasData",
)



import asyncio
import enum
import itertools

from telegrinder import (
    API,
    MESSAGE_FROM_USER,
    Message,
    MessageReplyHandler,
    Telegrinder,
    Token,
    WaiterMachine,
)
from telegrinder.modules import logger
from telegrinder.rules import EnumTextRule, StartCommand

api = API(token=Token.from_env())
bot = Telegrinder(api)
wm = WaiterMachine(bot.dispatch)

logger.set_level("INFO")


class YesOrNo(enum.Enum):
    YES = "Yes"
    NO = "No"


@bot.on.message(StartCommand())
async def start(message: Message) -> str:
    await message.answer("Do you want some tee?")
    _, ctx = await wm.wait(
        MESSAGE_FROM_USER,
        message.from_user.id,
        release=EnumTextRule(YesOrNo),
        on_miss=MessageReplyHandler("You want, dont you?", as_reply=True),
    )
    if ctx.enum_text == YesOrNo.NO:
        return "Leee thats sad dat tee is so sweet"

    await message.answer("Yay here is you tee with a nice krendeliok")

    m = (await message.answer("Preparing..")).unwrap()
    queue = itertools.cycle(["🫖🥨", "🥨🤗", "🤗🫖"])
    for _ in range(10):
        await m.edit(next(queue))
        await asyncio.sleep(0.5)

    return "Tee session is over! Goodbye!"


bot.run_forever()



import typing

from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.dispatch.return_manager.abc import BaseReturnManager, register_manager


class CallbackQueryReturnManager(BaseReturnManager):
    @register_manager(str)
    @staticmethod
    async def str_manager(value: str, event: CallbackQueryCute) -> None:
        await event.answer(value)

    @register_manager(dict)
    @staticmethod
    async def dict_manager(value: dict[str, typing.Any], event: CallbackQueryCute) -> None:
        await event.answer(**value)


__all__ = ("CallbackQueryReturnManager",)



import pytest

from telegrinder.types.objects import Update

UPDATE = Update.from_raw(
    b"""
    {
        "update_id":12345,
        "callback_query":{
            "id":"4382bfdwdsb323b2d9",
            "from":{
                "last_name":"Test Lastname",
                "type":"private",
                "id":1111111,
                "is_bot":false,
                "first_name":"Test Firstname",
                "username":"Testusername"
            },
            "chat_instance":"23asinstance23442",
            "data":"Data from button callback",
            "inline_message_id":"1234csdbsk4839",
            "message":{
                "message_id": 23423522,
                "from":{
                    "last_name":"Test Lastname",
                    "type": "private",
                    "id":1111111,
                    "is_bot":false,
                    "first_name":"Test Firstname",
                    "username":"Testusername"
                },
                "chat": {
                    "last_name":"Test Lastname",
                    "type": "private",
                    "id":1111111,
                    "is_bot":false,
                    "first_name":"Test Firstname",
                    "username":"Testusername"
                },
                "date":1234567898,
                "text":"Hello, Telegrinder! btw, laurelang - nice pure logical programming launguage ^_^"
            }
        }
    }
    """
)


@pytest.fixture()
def callback_query_update():
    return UPDATE



import typing

from telegrinder.node.base import ComposeError, NodeClass, scalar_node
from telegrinder.node.scope import per_call
from telegrinder.tools.magic.annotations import TypeParameter, get_generic_parameters


@scalar_node
@per_call
class TypeArgs:
    @classmethod
    def compose(cls, node_cls: NodeClass) -> dict[TypeParameter, typing.Any]:
        return get_generic_parameters(node_cls).expect(ComposeError("No generic alias."))


__all__ = ("TypeArgs",)



import abc
import dataclasses
import gettext
import os
import pathlib
import typing
from functools import cached_property

from telegrinder.node.base import DataNode, GlobalNode
from telegrinder.node.source import Locale

type Separator = KeySeparator

DEFAULT_SEPARATOR: typing.Final[str] = "-"


@dataclasses.dataclass(frozen=True)
class KeySeparator(GlobalNode[Separator], DataNode):
    value: str

    @classmethod
    def set(cls, value: str, /) -> None:
        super().set(cls(value))

    @classmethod
    def compose(cls) -> Separator:
        return cls.get(default=cls(DEFAULT_SEPARATOR))


@dataclasses.dataclass(kw_only=True)
class ABCTranslator(DataNode, abc.ABC):
    locale: str
    separator: str
    _keys: list[str] = dataclasses.field(default_factory=list[str], init=False)

    @typing.overload
    def __call__(self, **context: typing.Any) -> str: ...

    @typing.overload
    def __call__(self, message_id: str, /, **context: typing.Any) -> str: ...

    def __call__(self, message_id: str | None = None, **context: typing.Any) -> str:
        result = self.translate(message_id or self.message_id, **context)
        if not message_id:
            self._keys.clear()
        return result

    def __getattr__(self, __key: str) -> typing.Self:
        self._keys.append(__key)
        return self

    @property
    def message_id(self) -> str:
        return self.separator.join(self._keys)

    @abc.abstractmethod
    def translate(self, message_id: str, **context: typing.Any) -> str:
        pass

    @classmethod
    def compose(cls, locale: Locale, separator: KeySeparator) -> typing.Self:
        return cls(locale=locale, separator=separator.value)


@dataclasses.dataclass
class I18NConfig:
    domain: str
    folder: str | pathlib.Path
    default_locale: str = dataclasses.field(default="en")

    @cached_property
    def translators(self) -> dict[str, gettext.GNUTranslations]:
        result = {}

        for name in os.listdir(self.folder):
            if not os.path.isdir(os.path.join(self.folder, name)):
                continue

            mo_path = os.path.join(self.folder, name, "LC_MESSAGES", f"{self.domain}.mo")

            if os.path.exists(mo_path):
                with open(mo_path, "rb") as f:
                    result[name] = gettext.GNUTranslations(f)
            elif os.path.exists(mo_path[:-2] + "po"):
                raise FileNotFoundError(".po files should be compiled first")

        return result

    def get_translator(self, locale: str, /) -> gettext.GNUTranslations:
        locale = locale if locale in self.translators else self.default_locale
        return self.translators[locale]


class BaseTranslator(ABCTranslator):
    config: typing.ClassVar[I18NConfig]

    def __class_getitem__(cls, config: I18NConfig, /) -> typing.Any:
        return type(cls.__name__, (cls,), dict(config=config))

    @classmethod
    def configure(cls, config: I18NConfig, /) -> None:
        cls.config = config

    def translate(self, message_id: str, **context: typing.Any) -> str:
        return self.config.get_translator(self.locale).gettext(message_id).format(**context)


__all__ = ("ABCTranslator", "BaseTranslator", "I18NConfig", "KeySeparator")



from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Argument, Command

api = API(token=Token.from_env())
bot = Telegrinder(api)
logger.set_level("INFO")


def character(c: str) -> str | None:
    if len(c) != 1:
        return None
    return c


def sentence(s: str) -> str | None:
    if not s.startswith("'") and s.endswith("'"):
        return None
    s = s.removeprefix("'").removesuffix("'")
    return s


def int_validator(s: str) -> int | None:
    if not s.isdigit():
        return None
    return int(s)


# Try:
# /split 'hello,_my_friend' _ 1
# /split 'hello,_my_friend' _
# /split 'hello, my friend'   2
# /split 'hello, my friend' ,
@bot.on.message(
    Command(
        "split",
        Argument("string", [sentence]),
        Argument("sep", [character], optional=True),
        Argument("count", [int_validator], optional=True),
    )
)
async def split_handler(
    message: Message,
    string: str,
    sep: str = " ",
    count: int | None = None,
) -> None:
    await message.answer(" | ".join(string.split(sep, count) if count is not None else string.split(sep)))


@bot.on.message(
    Command(
        ["sum", "s"],
        Argument("x", [int_validator]),
        Argument("y", [int_validator]),
        Argument("r", [int_validator], optional=True),
    )
)
async def sum_handler(message: Message, x: int, y: int, r: int | None = None) -> str:
    if r is not None:
        return "Yes" if x + y == r else "No"
    return str(x + y)


bot.run_forever()



from examples.blueprint_bot.client import wm
from telegrinder import MESSAGE_FROM_USER, Dispatch, Message, MessageReplyHandler
from telegrinder.modules import logger
from telegrinder.rules import ABCRule, HasText, Markup, RuleEnum, Text

dp = Dispatch()


class Commands(RuleEnum):
    PLAY = Text(["/play", "/game"])
    GET_PRESENTS = Text("/get_presents")
    COMMIT_SUICIDE = Markup(["/suicide", "/suicide <reason>"])


class WasNaughty(ABCRule, requires=[HasText()]):
    async def check(self, message: Message) -> bool:
        await message.answer("Ok, but were you naughty this year?")
        m, _ = await wm.wait(
            MESSAGE_FROM_USER,
            message.from_user.id,
            release=Text(["yes", "no"], ignore_case=True),
            on_miss=MessageReplyHandler("Yes or no? Were you naughty??"),
        )
        return m.text.unwrap().lower() == "yes"


# The enumeration will only resolve once and save the state in context
dp.message.auto_rules = Commands()


@dp.message(Commands.PLAY)
async def handle_play(m: Message):
    logger.info("Playing ^_^")
    await m.answer("What a wonderful play..")


@dp.message(Commands.GET_PRESENTS, WasNaughty())
async def handler_get_presents_if_naughty(m: Message):
    await m.answer(
        "Well, well, well. You were naughty sooo i don't think i can give you some presents.. Only this dubious tomato for now"
    )
    await m.answer("🍅")
    await m.answer("Behave better next year :)")


@dp.message(Commands.GET_PRESENTS)
async def handle_get_presents(m: Message):
    await m.answer("Ho-ho-ho! These are your high quality presents:")
    await m.answer("🥐 (medialuna from argentina)")
    await m.answer("⚙️ wery important detail to fix your life")
    await m.answer("See ya next year!!")


@dp.message(Commands.COMMIT_SUICIDE)
async def handle_commit_suicide(m: Message, reason: str | None = None):
    if reason is None:
        await m.answer("Please specify your reason to do that")
        await wm.wait(MESSAGE_FROM_USER, m.from_user.id, release=HasText())
    await m.answer(
        "Thats a bad reason because nothing in life is final until it's over you know.. There is no reason to live but at the moment you are not knowing the reason you are living despite everything"
    )



import typing

from telegrinder.msgspec_utils import encoder


def encode_form_data(
    data: dict[str, typing.Any],
    files: dict[str, tuple[str, typing.Any]],
    /,
) -> dict[str, str]:
    context = dict(files=files)
    return {
        k: encoder.encode(v, context=context).strip('"')  # Remove quoted string
        if not isinstance(v, str)
        else v
        for k, v in data.items()
    }


class MultipartFormProto(typing.Protocol):
    def add_field(
        self,
        name: str,
        value: typing.Any,
        /,
        *,
        filename: str | None = None,
    ) -> None: ...


__all__ = ("MultipartFormProto", "encode_form_data")



import types
import typing

import msgspec
from fntypes.co import Nothing, Variative

if typing.TYPE_CHECKING:
    from telegrinder.tools.fullname import fullname
    from telegrinder.tools.magic.function import bundle

    def get_class_annotations(obj: typing.Any, /) -> dict[str, typing.Any]: ...

    def get_type_hints(obj: typing.Any, /) -> dict[str, typing.Any]: ...

else:
    from msgspec._utils import get_class_annotations, get_type_hints

    def bundle(*args, **kwargs):
        from telegrinder.tools.magic.function import bundle

        return bundle(*args, **kwargs)

    def fullname(*args, **kwargs):
        from telegrinder.tools.fullname import fullname

        return fullname(*args, **kwargs)


_COMMON_TYPES = frozenset((str, int, float, bool, None, Variative))


def get_origin[T](t: T, /) -> type[T]:
    t_ = typing.get_origin(t) or t
    t_ = type(t_) if not isinstance(t_, type) else t_
    return typing.cast("type[T]", t_)


def is_common_type[T](t: T, /) -> typing.TypeGuard[type[T]]:
    if not isinstance(t, type):
        return False
    return t in _COMMON_TYPES or issubclass(t, msgspec.Struct) or hasattr(t, "__dataclass_fields__")


def struct_asdict(
    struct: msgspec.Struct,
    /,
    *,
    exclude_unset: bool = True,
    unset_as_nothing: bool = False,
) -> dict[str, typing.Any]:
    return {
        k: v if not unset_as_nothing else Nothing() if v is msgspec.UNSET else v
        for k, v in msgspec.structs.asdict(struct).items()
        if not (exclude_unset and isinstance(v, msgspec.UnsetType | types.NoneType | Nothing))
    }


def type_check(obj: typing.Any, t: typing.Any) -> bool:
    return (
        isinstance(obj, t)
        if isinstance(t, type) and issubclass(t, msgspec.Struct)
        else type(obj) in t
        if isinstance(t, tuple)
        else type(obj) is t
    )


__all__ = (
    "get_class_annotations",
    "get_origin",
    "get_type_hints",
    "is_common_type",
    "struct_asdict",
    "type_check",
)
